import pandas as pd
import matplotlib.pyplot as plt


def _draw_hyper_param_evolution_build_data(grid_search_results, hyper_param, greater_is_better):
    # Get the indexes of the current hyper param - with the other hyper params being constant
    grid_search_results_df = pd.DataFrame(grid_search_results.cv_results_)
    best_grid_item = grid_search_results_df[grid_search_results_df['rank_test_score'] == 1]['params'].values[0].copy()
    del best_grid_item[hyper_param]
    grid_search_results_df['best_other_params'] = grid_search_results_df['params'].apply(
        lambda y: all([y[key] == best_grid_item[key] for key in best_grid_item]))
    results_df = grid_search_results_df[grid_search_results_df['best_other_params'] == 1].copy()

    for sample, style, plotly_style in (('train', '--', 'dash'), ('test', '-', 'solid')):
        try:
            results_df.loc[:, 'mean_spread_low_%s' % sample] = \
                results_df['mean_%s_score' % sample] - results_df['std_%s_score' % sample]
            results_df['mean_spread_high_%s' % sample] = \
                results_df['mean_%s_score' % sample] + results_df['std_%s_score' % sample]

            results_df.sort_values(by=['param_' + hyper_param], inplace=True)

        except ValueError:
            print("INFO: Hyperparam {0} cannot be displayed. It is not a quantitative param".format(hyper_param))

    if greater_is_better:
        best_param = results_df['param_' + hyper_param].loc[results_df['mean_test_score'].idxmax(axis=0)]
        best_score = results_df['mean_test_score'].max()
    else:
        best_param = results_df['param_' + hyper_param].loc[results_df['mean_test_score'].idxmin(axis=0)]
        best_score = results_df['mean_test_score'].min()

    score = 'score'
    try:
        score = grid_search_results.estimator.criterion
    except:
        pass

    return results_df, best_score, best_param, score


def _draw_hyper_param_evolution(grid_search_results, hyperparam, greater_is_better):
    results_df, best_score, best_param, score = _draw_hyper_param_evolution_build_data(
        grid_search_results, hyperparam, greater_is_better)
    fig, ax = plt.subplots(1, 1)

    # plot:
    color = 'g'
    for sample, style in (('train', '--'), ('test', '-')):
        ax.fill_between(list(results_df['param_' + hyperparam].values),
                        list(results_df['mean_spread_low_%s' % sample].values),
                        list(results_df['mean_spread_high_%s' % sample].values),
                        alpha=0.1 if sample == 'test' else 0, color=color)

        ax.plot(list(results_df['param_' + hyperparam].values),
                list(results_df['mean_%s_score' % sample].values),
                style, color=color, alpha=1 if sample == 'test' else 0.7, label="score (%s)" % sample)

    # dotted vertical line at the best score for that scorer marked by x and annotate the best score for that scorer
    ax.plot([best_param], [best_score], linestyle='None', color=color, marker='x', markeredgewidth=3, ms=8)
    ax.plot([best_param, best_param], [0, best_score], color=color, linestyle='-.')
    ax.annotate("%0.2f" % best_score, (best_param, best_score + 0.005))
    ax.legend(loc="best")
    ax.grid(False)

    # annotate the figure:
    ax.set_title('Hyperparam_evolution', fontdict={'fontsize': 8})
    ax.set_xlabel(hyperparam)
    ax.set_ylabel(score)
    return fig, ax


def score_evolution_hyper_params(grid_search_results, hyper_params, greater_is_better, search_type):
    """
    Evolution of the scoring according to the hyper parameters values
    :param grid_search_results: results from the grid_search
    :param hyper_params: dictionary of the hyper_params passed to the grid search
    :param greater_is_better: whether low score is better
    :param search_type: whether grid or randomized. If the search is randomized, the visualization won't be drawn
    :return: Plots
    """
    if search_type == 'grid':
        figs = []
        keys = [k for k in hyper_params.keys() if len(hyper_params[k]) > 1 and not isinstance(hyper_params[k][0], str)
                and not isinstance(hyper_params[k][0], tuple)]
        if len(keys) > 0:
            for index, hyperparam in enumerate(keys):
                figs.append(_draw_hyper_param_evolution(grid_search_results, hyperparam, greater_is_better))
            return [fig[0] for fig in figs]
    else:
        print("INFO: No grid search has been run. "
              "Please, configure a grid search to analyse the hyper_params evolution")
        return None


def visualization_methods(method):
    """
    This function shows all the visualization that are available:
        - score_evolution_hyper_params
    """
    return {
        'score_evolution_hyper_params': score_evolution_hyper_params
    }.get(method)



from mlapp.handlers.instance import database_handler, file_storage_handler
from mlapp.managers import DataManager, pipeline
import pandas as pd
import os
import numpy as np


class AdvancedRegressionDataManager(DataManager):
    # -------------------------------------- train methods -------------------------------------------
    @pipeline
    def load_train_data(self, *args):
        return self._load_data_inner(self.data_settings.get('data_sources', []))

    @pipeline
    def clean_train_data(self, data):
        missing_values = {}
        data_handling = self.data_settings.get('data_handling', {})
        features_handling = self.data_settings.get('features_handling', {})
        null_percentage = data_handling.get("feature_remove_by_null_percentage", 0.5)
        data = data.loc[:, data.isnull().mean() < null_percentage]

        for feature in features_handling.keys():
            missing_values[feature] = eval(features_handling[feature].get("fillna", "np.mean"))(data[feature].values)

        default_missing_features = list(set(data.columns).difference(set(list(features_handling.keys()))))
        default_missing_values = data[default_missing_features].mean(axis=0).to_dict()
        missing_values.update(default_missing_values)
        self.save_metadata('missing_values', missing_values)

        data = data.fillna(missing_values)
        return data

    @pipeline
    def transform_train_data(self, data):
        return self._transform_data(data)

    # ------------------------------------- forecast methods -----------------------------------------
    @pipeline
    def load_forecast_data(self, *args):
        # read data for forecasting
        data = self._load_data_inner(self.data_settings.get('data_sources', []))

        # creating prediction data for example
        if self.data_settings.get('generate_forecast_data') == 'index_based':
            idx = self.data_settings.get('generate_forecast_data_indices', data.index)
            data = data.loc[idx]
            data.drop([self.model_settings['variable_to_predict']], axis=1, inplace=True)
        else:
            data.drop([self.model_settings['variable_to_predict']], axis=1, inplace=True)
            data = pd.DataFrame(
                data=np.random.normal(data.mean().mean(), data.std().mean(), data.shape)[0:10], columns=data.columns)
        return data
    
    @pipeline
    def clean_forecast_data(self, data):
        return data

    @pipeline
    def transform_forecast_data(self, data):
        return self._transform_data(data)

    # ------------------------- load helper function for train/forecast ------------------------------
    def _load_data_inner(self, sources):
        data = pd.DataFrame()
        for source_type in sources.keys():
            source_data = pd.DataFrame()

            # local file
            if source_type == 'local':
                for file_path in sources[source_type]['file_paths']:
                    local_path = os.path.join(os.getcwd(), file_path)
                    source_data = pd.concat([source_data, self._load_data_from_file(local_path)])

            # database
            if source_type == 'db':
                source_data = self._load_data_from_db(sources[source_type]['query'], sources[source_type]['args'])

            # s3 file store
            if source_type == 's3':
                for bucket in sources[source_type]['buckets']:
                    for file_key in sources[source_type]['buckets'][bucket]['file_keys']:
                        source_local_file = self._load_file_from_file_store(bucket, file_key)
                        source_data = pd.concat([source_data, pd.read_csv(source_local_file)])

                        # removing local file
                        os.remove(source_local_file)

            data = pd.concat([data, source_data])

        return data

    @staticmethod
    def _load_data_from_file(path):
        return pd.read_csv(path, encoding='ISO-8859-1')

    @staticmethod
    def _load_data_from_db(query, params=None):
        if params is None:
            params = []

        return database_handler('POSTGRES').get_df(query, params)

    def _load_file_from_file_store(self, bucket, file_key):
        unicode_file_name = str(file_key)
        to_path = os.path.join(self.local_storage_path, unicode_file_name)
        file_storage_handler('MINIO').download_file(bucket, unicode_file_name, to_path)
        return to_path

    def _transform_data(self,data):
        data_handling = self.data_settings.get('data_handling', {})

        # interactions
        if data_handling.get('interactions', False):
            columns_list = [c for c in list(data.columns) if c != 'target']
            for col1 in columns_list:
                for col2 in columns_list:
                    if col1 != col2:
                        name = str(col1) + '_' + str(col2)
                        reverse_name = str(col2) + '_' + str(col1)
                        if reverse_name not in list(data.columns):
                            data[name] = (data[col1] + 1) * (data[col2] + 1)

        # binning
        for feature_to_bin in data_handling.get("features_to_bin", []):
            full_bins = [data[feature_to_bin['name']].min() - 1] + \
                        feature_to_bin['bins'] + [data[feature_to_bin['name']].max() + 1]

            data[feature_to_bin['name'] + '_binned'] = pd.cut(
                data[feature_to_bin['name']],
                bins=full_bins,
                labels=range(len(full_bins) - 1)).astype(float)

        # transformation
        for col in data_handling.get("features_handling", {}).keys():
            transformation_array = data_handling["features_handling"][col].get("transformation", [])
            # applying transformations
            for feature_transformation_method in transformation_array:
                data[col + '_' + feature_transformation_method] = eval(feature_transformation_method)(
                    data[col])

        # dropping features
        features_to_remove = data_handling.get('features_to_remove', [])
        if len(features_to_remove) > 0:
            data = data.drop([feature for feature in features_to_remove if feature in data.columns], axis=1)
        return data

    @pipeline
    def load_actuals_data(self, *args):
        index_to_y_true = self.model_settings.get('index_to_y_true', [])
        df = pd.DataFrame(index_to_y_true, index=[0])
        return df

    @pipeline
    def load_target_data(self, *args):
        raise NotImplementedError()


from azureml.core import Run
from mlapp.main import MLApp
from mlapp.handlers.wrappers.file_storage_wrapper import file_storage_instance
from mlapp.integrations.aml.utils.run_class import load_config_from_string, tag_and_log_run, tag_and_log_outputs
import argparse
from config import settings

from mlapp.managers.flow_manager import FlowManager

parser = argparse.ArgumentParser()
parser.add_argument('--config', type=str, dest='config', help='configuration')
args = parser.parse_args()
run = Run.get_context()

# pre-processing
config = load_config_from_string(args.config)
tag_and_log_run(config)

# init mlapp
MLApp(settings)

# run config
_, output_ids, output_data = FlowManager(Run.get_context().id, config).run()

# post-processing
tag_and_log_outputs(output_ids)

# post-processing
file_storage_instance.postprocessing()



rename_dictionary = {
    "base": {
        # key -> file_name: value -> list of dictionaries of words to replace.
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_train_config.json': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_forecast_config.json': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        }
    },
    "classification": {
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'},
                      {"word": '_feature_engineering', "word_format": 'str.lower', 'word_type': 'append-left'},
                      {"word": 'from assets.', "word_format": 'str.lower', 'word_type': 'append-right'},
                      {"word": 'FeatureEngineering', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'},
                      {"word": '_visualizations', "word_format": 'str.lower', 'word_type': 'append-left'},
                      {"word": 'from assets.', "word_format": 'str.lower', 'word_type': 'append-right'}]
        },
        '_train_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_forecast_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*\n*\s*:\s*\n*\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_feature_engineering_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*\n*\s*:\s*\n*\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_reuse_features_and_train_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*\n*\s*:\s*\n*\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_feature_engineering.py': {
            "words": [{"word": 'FeatureEngineering', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_visualizations.py': {
            "words": []
        }
    },
    "crash_course": {
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_train_config.json': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_forecast_config.json': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        }
    },
    "flow_regression": {
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_forecast_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        }
    },
    "basic_regression": {
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_train_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_forecast_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_custom_pipeline_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_feature_engineering_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_train_step_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_reuse_features_and_train_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        }
    },
    "spark_classification": {
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'},
                      {"word": '_feature_engineering', "word_format": 'str.lower', 'word_type': 'append-left'},
                      {"word": 'from assets.', "word_format": 'str.lower', 'word_type': 'append-right'},
                      {"word": 'FeatureEngineering', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_train_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_forecast_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_feature_engineering.py': {
            "words": [{"word": 'FeatureEngineering', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        }
    },
    "spark_regression": {
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_train_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_forecast_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
    },
    "advanced_regression": {
        '_data_manager.py': {
            "words": [{"word": 'DataManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_model_manager.py': {
            "words": [{"word": 'ModelManager', "word_format": 'str_capitalize', 'word_type': 'append-left'}]
        },
        '_train_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_forecast_config.py': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        },
        '_prediction_accuracy_config.json': {
            'inner_path': 'configs',
            "words": [{"word": '"asset_name": "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*"'},
                      {"word": '"asset_name": \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'"asset_name"\s*:\s*\''},
                      {"word": '\'asset_name\': \'', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*\''},
                      {"word": '\'asset_name\': "', "word_format": 'str.lower', 'word_type': 'append-right',
                       "word_pattern": r'\'asset_name\'\s*:\s*"'},
                      {"word": '_config ', "word_format": 'str.lower', 'word_type': 'append-left'}]
        }
    }
}



import numpy as np
import os
from mlapp.handlers.instance import spark_handler
from assets.spark_classification.spark_classification_feature_engineering import SparkClassificationFeatureEngineering
from mlapp.managers import DataManager, pipeline
from mlapp.utils.exceptions.base_exceptions import ConfigKeyError


class SparkClassificationDataManager(DataManager):
    # -------------------------------- custom init function ---------------------------------------
    def __init__(self, config, *args, **kwargs):
        DataManager.__init__(self, config, *args, **kwargs)

        # Custom initiates
        if self.data_settings.get('data_handling', None) is None:
            raise ConfigKeyError("'data_handling' is needed in model configuration file!!!")

        # Feature engineering helper class
        self.feature_engineering_instance = SparkClassificationFeatureEngineering()

    # -------------------------------------- train methods -------------------------------------------
    @pipeline
    def load_train_data(self,*args):
        return self._load_data()

    @pipeline
    def clean_train_data(self, data):
        print('------------- CLEAN TRAIN DATA -------------')
        data, missing_values = self._clean_data(data)
        self.save_metadata('missing_values', missing_values)
        return data

    @pipeline
    def transform_train_data(self, data):
        results_df = self._transform_data(data)
        print("------------------------- return final features to train -------------------------")
        return results_df

    # ------------------------------------- forecast methods -----------------------------------------
    @pipeline
    def load_forecast_data(self, *args):
        return self._load_data()

    @pipeline
    def clean_forecast_data(self, data):
        print('------------- CLEAN FORECAST DATA -------------')
        data, missing_values = self._clean_data(data)
        return data

    @pipeline
    def transform_forecast_data(self, data):
        data_handling = self.data_settings.get('data_handling', {})
        dates_transformation = data_handling.get("dates_transformation", {})
        return self._transform_data(data, dates_transformation=dates_transformation)

    def transform_data_for_exploration(self, data):
        data_handling = self.data_settings.get('data_handling', {})

        for col in data.columns:
            if data[col].dtype == np.bool:
                data[col] = data[col].astype(int)

        data_df = self.feature_engineering_instance.transform_features(data, data_handling.get('features_handling'))

        print("------------------------- handling y variable -------------------------")
        final_y = data_df[self.data_settings["variable_to_predict"]]
        data_df = data_df.drop(self.data_settings["variable_to_predict"], axis=1)

        print("------------------------- merge final features with y value -------------------------")
        results_df = data_df.join(final_y.rename(self.data_settings["variable_to_predict"]))

        print("------------------------- return final features to train -------------------------")
        return results_df

    # ------------------------------ helper methods for train/forecast --------------------------------
    def _load_data(self):
        local_path = os.path.join(os.getcwd(), self.data_settings["local_data_csvs"][0].get("path", ""))
        return spark_handler('LOCAL-SPARK').load_csv_file(local_path, inferSchema=True)

    def _clean_data(self, data):
        data_handling = self.data_settings.get('data_handling')
        set_features_index = data_handling.get('set_features_index', [])
        if set_features_index:
            data = data.set_index(set_features_index)

        features_for_train = data_handling.get('features_for_train', [])
        features_to_remove = data_handling.get('features_to_remove', [])

        if features_to_remove:
            for feature_to_remove in features_to_remove:
                if feature_to_remove in data.columns and (
                        feature_to_remove != self.data_settings["variable_to_predict"]):
                    data = data.drop(feature_to_remove)

        if features_for_train:
            features_for_train = list(filter(lambda x: x not in features_to_remove, features_for_train))
            features_for_train += [self.data_settings["variable_to_predict"]]
            data = data[features_for_train]

        print("------------------------- Removing high percentage of null features -------------------------")
        data = self.feature_engineering_instance.remove_features_by_null_threshold(data, data_handling.get(
            'feature_remove_by_null_percentage', 0.3))

        print("------------------------- fill na -------------------------")
        stored_missing_values = self.get_metadata('missing_values', default_value={})

        data, missing_values = self.feature_engineering_instance.fillna_features(
            data, data_handling.get('features_handling', {}), data_handling.get('default_missing_value', 0),
            stored_missing_values)

        return data, missing_values

    def _transform_data(self, data_df, dates_transformation=None):
        data_handling = self.data_settings.get('data_handling', {})
        # Convert boolean columns to int:
        data_df = data_df.select(
            [data_df[colName].cast('int') if colType == 'boolean' else data_df[colName] for colName, colType in data_df.dtypes])

        data_df = self.feature_engineering_instance.transform_features(data_df, data_handling.get('features_handling'))
        dates_transformation = data_handling.get('dates_transformation', {})
        if dates_transformation:
            for col in dates_transformation.get('columns', []):
                pass
                # TODO:
                # data_df = self.feature_engineering_instance.get_days_from_date(
                #     data_df, col,
                #     dates_transformation.get('extraction_date') if extraction_date is None else extraction_date)

        print("------------------------- handling y variable -------------------------")
        data_df = self.feature_engineering_instance.handle_y_variable(
            data_df, self.data_settings["variable_to_predict"], data_handling['y_variable'])

        print("------------------------- feature interactions -------------------------")
        interactions_list = data_handling.get('features_interactions', [])
        data_df = self.feature_engineering_instance.interact_features(data_df, interactions_list)

        print("---------------------- Convert dates to days elapsed from date -------------")
        data_df = self.feature_engineering_instance.transform_dates_to_elapsed_days(
            data_df, data_handling.get("dates_format", 'yyyyMMdd'))

        print("------------------------- bin continuous features -------------------------")
        data_df = self.feature_engineering_instance.bin_continuous_features(
            data_df, data_handling.get("features_to_bin", []))

        print("---------------------- transform categorical to dummy variables -------------")
        data_df = self.feature_engineering_instance.convert_features_to_dummies(
            data_df, data_handling.get("features_to_convert_to_dummies", []))

        data_df.cache()
        return data_df

    def load_target_data(self, *args):
        raise NotImplementedError()




import os
import shutil
import imghdr
from mlapp.handlers.file_storages.file_storage_interface import FileStorageInterface
from mlapp.handlers.databases.postgres_handler import PostgresHandler
from mlapp.integrations.aml.utils.constants import OUTPUTS_FOLDER, AML_MLAPP_FOLDER
from mlapp.utils.general import read_json_file
from azureml.core import Run, Experiment
from azureml.exceptions import RunEnvironmentException
from mlapp.utils.exceptions.framework_exceptions import SkipServiceException

# postgresql queries
RUN_QUERY = '''SELECT * FROM map_run_id WHERE run_id={0}'''
UPDATE_RUN_QUERY = '''INSERT INTO map_run_id (aml_run_id, run_id, experiment_name) VALUES ({0}, {1}, {2})'''


class AmlRunStorageHandler(FileStorageInterface):
    def __init__(self, settings):
        """
        Initializes the AmlRunStorageHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(AmlRunStorageHandler, self).__init__()
        self.temporary_storage = settings.get('temporary_storage_path')

        self.imgs_bucket = settings.get('file_store_buckets', {}).get('imgs', 'mlapp-imgs')
        self.metadata_bucket = settings.get('file_store_buckets', {}).get('metadata', 'mlapp-metadata')

        # getting azureml workspace
        try:
            self.global_run = Run.get_context(allow_offline=False)
        except RunEnvironmentException as e:
            raise SkipServiceException('Skip AmlModelStorageHandler handler')
        self.ws = self.global_run.experiment.workspace

        # creating postgres database handler
        self.db_handler = PostgresHandler(settings)

        # creating run_ids to save for postprocessing
        self.run_ids = {}

    def _get_current_run(self, run_id):
        # get azureml Run and Experiment details from postgres db
        result = self.db_handler.execute_query(RUN_QUERY, params=[run_id])
        if len(result) == 0:
            raise Exception('ERROR: run context of run_id `%s` not exists, please make sure `%s` is the right id.' %
                            (run_id, run_id))
        db_aml_run_id, db_run_id, db_experiment_name = result.pop()

        # creating Run object
        experiment = Experiment(workspace=self.ws, name=db_experiment_name)
        run = Run(experiment=experiment, run_id=db_aml_run_id)
        return run

    def download_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        This method downloads a file into your machine from aml datastore
        :param bucket_name:  name of the bucket/container, unnecessary parameter
        :param object_name: name of the object/file
        :param file_path: path to local file, unnecessary parameter
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: None
        """
        # get run_id from object_name
        current_id = object_name.split('_')[0]

        # get current_id azureml Run object
        current_run: Run = self._get_current_run(current_id)

        # download file from current_run to file_path
        current_run.download_file(name=object_name, output_file_path=file_path)

    def upload_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        This method upload a file to the Run Object of AML
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: None
        """
        run_id = kwargs.get('run_id')
        if run_id not in self.run_ids:
            self.run_ids[run_id] = True

        # upload file to Run
        self.global_run.upload_file(name=object_name, path_or_stream=file_path)

        # log images
        if bucket_name == self.imgs_bucket:
            # check if file is an image
            is_img = imghdr.what(file_path)
            if is_img is not None:
                file_name = os.path.splitext(os.path.basename(file_path))[0]
                image_name = '_'.join(file_name.split('_')[1:])
                self.global_run.log_image(name=image_name, path=file_path)
        # log metadata
        else:
            if bucket_name == self.metadata_bucket:
                model_metadata = read_json_file(file_path)
                scores = model_metadata.get('models', {}).get('scores', None)
                if scores is not None:
                    for score_key, score_value in scores.items():
                        self.global_run.log(name=score_key, value=score_value)

    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        # get current_id azureml Run object
        current_run: Run = self._get_current_run(prefix)

        # return a list of files from the current run
        return current_run.get_file_names()

    def postprocess(self):
        """
        This method saves mapping from (run_id x experiment_name) to (run_id).
        :return: None
        """
        for run_id in self.run_ids.keys():
            self.db_handler.execute_query(UPDATE_RUN_QUERY, [self.global_run.id, run_id, self.global_run.experiment.name])



from azureml.core.runconfig import RunConfiguration
from azureml.core import ScriptRunConfig
import os
from azureml.core import Experiment


def run_local_compute_experiment(ws, experiment_name, entry_script, source_directory=os.getcwd()):
    # Edit a run configuration property on the fly.
    run_local = RunConfiguration()
    run_local.environment.python.user_managed_dependencies = True

    exp = Experiment(workspace=ws, name=experiment_name)

    src = ScriptRunConfig(source_directory=source_directory, script=entry_script, run_config=run_local)
    run = exp.submit(src)
    run.wait_for_completion(show_output=True)



import logging
from scipy import stats
import numpy as np
import pandas as pd
from sklearn.preprocessing import PolynomialFeatures
from pandas.core.dtypes.common import ensure_float64
from sklearn.metrics import precision_score, recall_score

logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)


def _create_evaluate_df_results(evaluations_binaries):
    results = pd.DataFrame()
    results['feature'] = evaluations_binaries.keys()
    results['feature_original_name'] = [evaluations_binaries[feature].get('feature_original_name')
                                        for feature in evaluations_binaries]
    results['baseline_precision'] = [evaluations_binaries[feature].get('baseline_precision')
                                     for feature in evaluations_binaries]
    results['count_1'] = [evaluations_binaries[feature].get('count_1_in_x')
                          for feature in evaluations_binaries]
    results['precision_by_1'] = [evaluations_binaries[feature].get('precision_by_1')
                                 for feature in evaluations_binaries]
    results['recall_by_1'] = [evaluations_binaries[feature].get('recall_by_1')
                              for feature in evaluations_binaries]
    results['count_0'] = [evaluations_binaries[feature].get('count_0_in_x')
                          for feature in evaluations_binaries]
    results['precision_by_0'] = [evaluations_binaries[feature].get('precision_by_0')
                                 for feature in evaluations_binaries]
    results['recall_by_0'] = [evaluations_binaries[feature].get('recall_by_0')
                              for feature in evaluations_binaries]
    return results


def evaluate_df_with_binary_output(X_df, y, pos_label=1, categorical_features=None, bins_dict=None):
    if not isinstance(X_df, pd.DataFrame):
        raise Exception('df argument should be pandas.DataFrame only')

    if not categorical_features:
        categorical_features = []
    if not bins_dict:
        bins_dict = {}

    if isinstance(y, pd.DataFrame):
        y = np.asarray(y[y.columns[0]])
    else:
        y = np.asarray(y)

    results = {}
    y_column = ensure_float64(y)
    i = 0
    for col in X_df.columns:
        is_categorical = True if col in categorical_features else False
        bins = []
        if col in bins_dict.keys():
            bins = bins_dict[col]
        c = _evaluate_binary_output(
            X_df[[col]], y_column, pos_label=pos_label, is_categorical=is_categorical, bins=bins)
        results.update(c)
        i += 1
    results_df = _create_evaluate_df_results(results)
    return results_df


def _evaluate_binary_output(x, y, pos_label=1, is_categorical=False, bins=None):
    try:
        if len(x.columns) > 1:
            raise Exception("need only one column in x")

        if not bins:
            bins = []

        result = {}
        feature_original_name = x.columns[0]

        if is_categorical:
            x[x.columns[0]] = x[x.columns[0]].astype(object)
            dummies = pd.get_dummies(x, prefix=x.columns[0], prefix_sep='_|_')
            for d in dummies.columns:
                result[d] = _evaluate_binary_feature(dummies[d], y, feature_original_name, pos_label)
            return result

        if len(x[x.columns[0]].unique()) == 1:
            for col in x.columns:
                result[col] = {}
            return result

        if len(x[x.columns[0]].unique()) == 2:
            x_max = max(x[x.columns[0]])
            x[x.columns[0]] = x[x.columns[0]].apply(lambda n: 1 if x_max == n else 0)

            for col in x.columns:
                result[col] = _evaluate_binary_feature(x, y, feature_original_name, pos_label)
            return result
        # continuous variable
        else:
            if len(bins) == 0:
                result[x.columns[0]] = {}
            else:
                cur_bins = bins
                labels = range(len(cur_bins) - 1)
                print(cur_bins)
                categorical_x = pd.DataFrame(data=pd.cut(x[x.columns[0]].apply(float), bins=cur_bins,
                                             labels=labels), columns=x.columns)
                result = _evaluate_binary_output(categorical_x, y, pos_label, is_categorical=True)
            return result
    except Exception as e:
        raise e


def _evaluate_binary_feature(x, y, feature_original_name, pos_label=1):
    x = ensure_float64(x)
    y = ensure_float64(y)
    if not len(np.unique(x)) <= 2 or not len(np.unique(y)):
        raise Exception("please provide binary x and y")
    y_positive_rate = np.count_nonzero(y == pos_label)/float(len(y))
    count_0 = np.count_nonzero(x == 0)
    count_1 = np.count_nonzero(x == 1)
    precision_0 = precision_score(y, 1.0 - x, pos_label=pos_label)
    recall_0 = recall_score(y, 1.0 - x, pos_label=pos_label)
    precision_1 = precision_score(y, x, pos_label=pos_label)
    recall_1 = recall_score(y, x, pos_label=pos_label)
    is_0_informative = precision_0 > y_positive_rate
    is_1_informative = precision_1 > y_positive_rate

    metrics = {'feature_original_name': feature_original_name,
               'baseline_precision': y_positive_rate,
               'count_0_in_x': count_0,
               'count_1_in_x': count_1,
               'precision_by_0' : precision_0,
               'recall_by_0' : recall_0,
               'precision_by_1' : precision_1,
               'recall_by_1' : recall_1,
               'is_0_informative': is_0_informative,
               'is_1_informative': is_1_informative}
    return metrics


def polynomial_features_labeled(raw_input_df, power):
    """
        This is a wrapper function for sklearn's Ploynomial features, which returns the resulting power-matrix with
        meaningful labels. It calls sklearn.preprocessing.PolynomialFeatures on raw_input_df with the specified power
        parameter and return it's resulting array as a labeled pandas dataframe (i.e cols= ['a', 'b', 'a^2', 'axb', 'b^2']).
        :param raw_input_df: labeled pandas dataframe.
        :param power: The degree of the polynomial features (use the same power as you want entered into
        PolynomialFeatures(power) directly).
        :return: power-matrix with meaningful column labels.
    """
    input_df = raw_input_df.copy()
    poly = PolynomialFeatures(power)
    output_nparray = poly.fit_transform(input_df)
    powers_nparray = poly.powers_

    input_feature_names = list(input_df.columns)
    target_feature_names = ["Constant Term"]
    for feature_distillation in powers_nparray[1:]:
        final_label = ""
        for i in range(len(input_feature_names)):
            if feature_distillation[i] == 0:
                continue
            else:
                variable = input_feature_names[i]
                power = feature_distillation[i]
                intermediary_label = "%s^%d" % (variable, power)
                if final_label == "":
                    final_label = intermediary_label
                else:
                    final_label = final_label + " x " + intermediary_label
        target_feature_names.append(final_label)
    output_df = pd.DataFrame(output_nparray, columns=target_feature_names, index=input_df.index)
    return output_df


def lag_feature(feature_df, lag, dropna=False):
    """
        Shifts feature forward by a specified lag. first n=lag values will be None.
        :param feature_df: A pandas series.
        :param lag: int. lag size.
        :param dropna: performs dropna after shift. default-False.
        :return: Transformed feature with the lag value added to the column name (i.e col_name_lag_5)..
    """
    res = pd.DataFrame(columns=['%s_lag_%s' % (feature_df.name, str(lag))],
                       data=feature_df.shift(lag, fill_value=None).values,
                       index=feature_df.index)
    if dropna:
        return res.dropna()
    else:
        return res


def lead_feature(feature_df, lead, dropna=False):
    """
      Shifts feature backward by a specified lead. last n=lead values will be None.
      :param feature_df: A pandas series.
      :param lead: int. lead size.
      :param dropna: performs dropna after shift. default-False.
      :return: Transformed feature with the lead value added to the column name (i.e col_name_lead_5).
    """
    res = pd.DataFrame(columns=['%s_lead_%s' % (feature_df.name, str(lead))],
                       data=feature_df.shift(-lead, fill_value=None).values,
                       index=feature_df.index)
    if dropna:
        return res.dropna()
    else:
        return res


def log_feature(feature_df):
    """
         Returns a natural log transformation of the feature.
         :param feature_df: A pandas series.
         :return: Transformed feature with 'log' added to the column name (i.e col_name_log).
    """
    feature_df_bis = feature_df.copy()
    try:
        feature_df_bis.rename('%s_log' % feature_df.name, inplace=True)
        return feature_df_bis.apply(np.log)
    except:
        logger.info('Not possible to log', feature_df.name)
        return feature_df


def exponent_feature(feature_df):
    """
         Returns a natural exponent transformation of the feature.
         :param feature_df: A pandas series.
         :return: Transformed feature with 'exp' added to the column name (i.e col_name_exp).
    """
    feature_df_bis = feature_df.copy()
    try:
        feature_df_bis.rename('%s_exp' % feature_df.name, inplace=True)
        return feature_df_bis.apply(np.exp)
    except:
        logger.info('Not possible to exponent', feature_df.name)
        return feature_df


def power_feature(feature_df, power):
    """
         Returns the feature raised to the power specified.
         :param feature_df: A pandas series.
         :param power: int.
         :return: Transformed feature with power value added to the column name (i.e col_name_pow_5).
    """
    feature_df_bis = feature_df.copy()
    try:
        feature_df_bis.rename('%s_pow_%s' % (feature_df.name, str(power)), inplace=True)
        return feature_df_bis.apply(lambda x: np.power(x, power))
    except:
        logger.info('Not possible to power {0}'.format(power), feature_df.name)
        return feature_df


def sqrt_feature(feature_df):
    """
        Returns a square root transformation of the feature.
        :param feature_df: A pandas series.
        :return: Transformed feature with 'sqrt' added to the column name (i.e col_name_sqrt).
   """
    feature_df_bis = feature_df.copy()
    try:
        feature_df_bis.rename('%s_sqrt' % feature_df.name, inplace=True)
        return feature_df_bis.apply(np.sqrt)
    except:
        logger.info('Not possible to sqrt', feature_df.name)
        return feature_df


def inverse_feature(feature_df):
    """
        Returns the inverse (1/x) transformation of the feature.
        :param feature_df: A pandas series.
        :return: Transformed feature with 'inverse' added to the column name (i.e col_name_inverse).
   """
    feature_df_bis = feature_df.copy()
    try:
        feature_df_bis.rename('%s_inverse' % feature_df.name, inplace=True)
        return feature_df_bis.apply(lambda x: float(1) / x)
    except:
        logger.info('Not possible to inverse', feature_df.name)
        return feature_df


def interact_features(feature_df, interact_list, drop_original_columns=True):
    """
    This function create interactions between pairs of features
    :param feature_df: a pandas dataframe
    :param interact_list: list of lists or tuples with two strings each, representing columns to be interacted.
    :param drop_original_columns: (bool) if set to True, columns to be interacted will be droped from the dataframe. note-
    if set to true a column cannot appear in more then one interaction pair.
    :return: DataFrame with the interactions columns, without the original columns.
    """

    for features_pair in interact_list:
        feature_0 = features_pair[0]
        feature_1 = features_pair[1]

        if (feature_0 not in feature_df.columns) or (feature_1 not in feature_df.columns):
            print('Warning: one the features: ' + feature_0 + ',' + feature_1 + 'do not exists in the Data.')
        else:
            new_feature = feature_df[feature_0] * feature_df[feature_1]
            if drop_original_columns:
                feature_df = feature_df.drop(feature_0, axis=1)
                feature_df = feature_df.drop(feature_1, axis=1)
            feature_df[feature_0 + '_X_' + feature_1] = new_feature

    return feature_df


def _remove_duplicate_columns(df):
    """
    find duplicate column names and keep only the first column.
    :param df: pandas dataframe.
    :return: pandas dataframe with duplicates removed.
    """
    cols = list(df.columns)
    to_drop = False
    for i, item in enumerate(df.columns):
        if item in df.columns[:i]:
            cols[i] = "toDROP"
            to_drop = True
    df.columns = cols
    if to_drop:
        df = df.drop("toDROP", 1)
    return df


def extend_dataframe(df, y_name_col=None, index_col=None, lead_order=3, lag_order=3, power_order=2, log=True, exp=True, sqrt=True, poly_degree=2, dropna =False, fillna_value=0, inverse=True):
    """
    Create a new dataframe with transformed features added as new columns.
    :param df: initial dataframe
    :param y_name_col: y column name. this column will not be extended.
    :param index_col: index column/s (string or list of strings). Any column that should NOT be extended, can be listed here.
    :param lead_order: shifts the feature values back. Extended dataframe will include all leads from 1 to lead_order specified.
    :param lag_order: shifts the feature values forward. Extended dataframe will include all lags from 1 to lead_order specified.
    :param power_order: Extended dataframe will include all powers from 1 to power_order specified.
    :param log: perform natural log transformation. Default=True.
    :param exp: perform natural exponent transformation. Default=True.
    :param sqrt: perform square root tranformation. Default=True.
    :param poly_degree: the highest order polynomial for the transformation. 0=no polynomial transformation.
    :param dropna: default True.
    :param fillna_value: default True.
    :param inverse: perform inverse transformation (1/x). default=True.
    :return: extended dataframe.
    """
    if isinstance(df, pd.Series):
        df = df.to_frame()

    not_to_extend = []
    if index_col is not None:
        if type(index_col) is list:
            not_to_extend += [ix_col for ix_col in index_col if ix_col in df.columns]
        elif index_col in df.columns:
            not_to_extend += [index_col]
    if y_name_col is not None and y_name_col in df.columns:
        not_to_extend += [y_name_col]

    features_columns = df.columns[~df.columns.isin(not_to_extend)]
    result = df[features_columns].copy()

    if isinstance(df, pd.DataFrame):
        for feature_name in features_columns:
            feature_df = df[feature_name]
            if lead_order > 0:
                for lead in range(1, lead_order + 1):
                    transformed_feature = lead_feature(feature_df, lead)
                    transformed_feature.rename(columns={feature_name: 'lead_' + feature_name}, inplace=True)
                    result = pd.concat([result, transformed_feature], axis=1)
            if lag_order > 0:
                for lag in range(1, min(lag_order, len(df)) + 1):
                    transformed_feature = lag_feature(feature_df, lag)
                    transformed_feature.rename(columns={feature_name: 'lag_' + feature_name}, inplace=True)
                    result = pd.concat([result, transformed_feature], axis=1)
            if power_order > 1:
                for power in range(2, power_order + 1):
                    transformed_feature = power_feature(feature_df, power)
                    result['power_' + feature_name] = transformed_feature
                    # transformed_feature.rename(columns={feature_name: 'power_' + feature_name}, inplace=True)
                    # result = pd.concat([result, transformed_feature], axis=1)
            if log > 0:
                if (feature_df >= 1).all():
                    transformed_feature = log_feature(feature_df)
                    result['log_' + feature_name] = transformed_feature
                    result.replace(-np.inf, 0, inplace=True)
            if exp > 0:
                if (feature_df < 100).all():
                    transformed_feature = exponent_feature(feature_df)
                    result['exp_' + feature_name] = transformed_feature
            if sqrt > 0:
                if (feature_df > 0).all():
                    transformed_feature = sqrt_feature(feature_df)
                    result['sqrt_' + feature_name] = transformed_feature
                    result.replace(-np.inf, 0, inplace=True)
            if inverse:
                if (feature_df != 0).all():
                    transformed_feature = inverse_feature(feature_df)
                    result['inverse_' + feature_name] = transformed_feature

        if poly_degree > 0:
            transformed_feature = polynomial_features_labeled(df[features_columns], poly_degree)
            result = pd.concat([result, transformed_feature], axis=1)
        final_extended_df = _remove_duplicate_columns(result)

        if dropna:
            final_extended_df.dropna(inplace=True)
        else:
            final_extended_df.fillna(fillna_value, inplace=True)

        if len(not_to_extend) > 0:
            final_extended_df = pd.concat([df[not_to_extend], final_extended_df], axis=1)

        return final_extended_df

    else:
        logger.error('Only works for DataFrame or Series')
        return None


def _calc_se(X, y, y_hat, DOF):
    """
    calc the formula: SE = sqrt(C.diagonal)
    C = MSE * (X'X)^-1 is the variance-covariance matrix
    MSE = SSE / DOF, Where  SSE = sum((y - y_hat)^2) and DF = n - k - 1 where
    n = number of obsevations, k = number of coefficients
    X = features matrix. X' = X transpose matrix
    @param X: Dataframe or Series
    @param y: Dataframe or Series
    @param y_hat: Dataframe or Series
    @param DOF: int - degree of freedom
    @return: numpay array
    """
    if type(y) is pd.DataFrame: y = y[y.columns[0]]
    SSE = np.sum(np.power(y - y_hat, 2))
    MSE = np.true_divide(SSE, DOF)
    X_transpose = pd.DataFrame.transpose(X)
    X_mul_X_transpose = np.dot(X_transpose.values, X.values)
    C = np.linalg.inv(X_mul_X_transpose)
    X_diagonal = np.diagonal(C)
    SE = pd.Series(np.sqrt(MSE * X_diagonal))
    return SE


def calc_t_values(X, y, y_hat, coefficients):
    """
    p-values for regerssion assets
    calc t values: t(b_i) = b_i / SE(b_i)
    Where b_i is the beta (coefficient) of x_i
    and SE(b_i) is the standard error of the coefficient
    @param X: Dataframe or Series
    @param y: Dataframe or Series
    @param y_hat: Dataframe or Series
    @param coefficients: list or ndarray
    @return: numpay array
    """
    if X.shape[1] != len(coefficients):
        raise Exception('X shape ' + repr(X.shape) + ' not match coefficients length ' + repr(len(coefficients)))

    # degree of freedom
    DOF = len(X) - len(coefficients) - 1
    if DOF < 1:
        raise Exception('Degrees Of Freedom must be greater or equals to 1')

    SE = _calc_se(X, y, y_hat, DOF)
    SE = pd.Series([1 if e == 0 else e for e in SE])
    t_statistics = np.true_divide(coefficients, SE)
    return t_statistics


def calc_p_values(X, y, y_hat, coefficients):
    """
    p-values for regression assets. Calc p values from t values
    @param X: pd.DataFrame or pd.Series
    @param y: pd.DataFrame or pd.Series
    @param y_hat: pd.DataFrame or pd.Series
    @param coefficients: list or ndarray
    @return: NumPy array
    """
    if X.shape[1] != len(coefficients):
        raise Exception('X shape ' + repr(X.shape) + ' not match coefficients length ' + repr(len(coefficients)))

    # degree of freedom
    DOF = len(X) - len(coefficients) - 1
    if DOF < 1:
        raise Exception('Degrees Of Freedom must be greater or equals to 1')

    t_statistics = calc_t_values(X, y, y_hat, coefficients)
    p_values = map(lambda t: 2 * stats.t.sf(t, DOF), np.abs(t_statistics))
    return list(p_values)



from environs import Env
from os.path import join

EMPTY_STRING = ""
ZERO = 0


class EnvironmentLoader:
    """ This class is responsible for load and filter environment variables. """

    @staticmethod
    def load(filename, path=''):
        """
        This function read env file
        :param filename: String, env filename
        :param path: String, env file path (Optional)
        :return: Env object
        """
        try:
            env = Env()
            full_path = join(path, filename)
            env.read_env(full_path, recurse=False, override=True)

            return env
        except ValueError:
            return Env()
        except Exception as e:
            raise Exception('Error in EnvironmentLoader, load method ' + str(e))

    @staticmethod
    def create_services(env, services_dict, available_services):
        """
        This file creates the services based on the `.env` file and environment variables.
        """
        services_settings = {}

        for service_name, service_type in services_dict.items():
            services_settings[service_name] = available_services(env)[service_type](service_name)

        return services_settings



import time
import datetime as dt
import importlib.util
import sys
import os
from mlapp.config import settings
from mlapp.utils.exceptions.base_exceptions import PipelineManagerException, FrameworkException
from mlapp.managers.io_manager import IOManager

AVAILABLE_STAGES = {}
BASE_CLASS_NAME = ''
TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
MANAGER_TYPES = {
    'data_manager': 'DataManager',
    'model_manager': 'ModelManager'
}


class PipelineManager(object):
    def __init__(self, run_id, pipeline_input, _input: IOManager, _output: IOManager, config, *args, **kwargs):
        """
        :param pipeline_input: the pipeline name string or list of strings
        :param _input: IOmanager instance with input to the pipeline
        :param _output: IOmanager instance to store the outputs of the pipelines to be saved externally
        :param config: config string of the pipeline
        :param args:
        :param kwargs:
        """
        for asset_name in AVAILABLE_STAGES:
            if asset_name != BASE_CLASS_NAME:
                AVAILABLE_STAGES[asset_name] = {}

        self.pipeline_name = ''

        # pipeline can be either list of stages or string of a default pipeline
        if isinstance(pipeline_input, list):
            self.stages = pipeline_input
        if isinstance(pipeline_input, str):
            self.pipeline_name = " '" + pipeline_input + "'"
            self.stages = settings.get('pipelines', {}).get(pipeline_input, [])

        self.config = config
        self.run_id = run_id
        self.input_manager = _input
        self.output_manager = _output
        self.asset_name = self.config.get('job_settings', {}).get('asset_name', '')

        self.data_manager_instance = self.create_manager_instance('data')
        self.model_manager_instance = self.create_manager_instance('model')

        # first inputs
        self.state = dict.fromkeys(self.stages, {})

    def create_manager_instance(self, manager_type):
        """
        Creates manager instance which class is defined in the asset. For example: model_manager or data_manader
        :param manager_type: the type : e.g. "data", "model"
        :return: instance of the manager
        """

        manager_file_name = self.asset_name + '_' + manager_type + '_manager'
        manager_module = 'assets.' + self.asset_name + '.' + manager_file_name
        manager_module_path = os.path.join('assets', self.asset_name, f'{manager_file_name }.py')
        manager_class_name = ''.join(x.capitalize() or '_' for x in manager_file_name.split('_'))  # CamelCase

        try:
            spec = importlib.util.spec_from_file_location(manager_module, manager_module_path)
            module = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = module
            spec.loader.exec_module(module)

            manager_class = getattr(module, manager_class_name)
            return manager_class(self.config.copy(), self.input_manager, self.output_manager, self.run_id)

        except Exception as e:
            print("Couldn't import class of " + manager_type + " manager for model: " + self.asset_name)
            print(">>> Please verify your " + manager_type +
                  " manager file/directory/class names are in the following convention:")
            print(">>> Directory: {{asset_name}}")
            print(">>> File: {{asset_name}}_" + manager_type + "_manager.py")
            print(">>> Class: {{asset_name_capitalized}}" + manager_type.capitalize() + "Manager")
            print(">>> Expected Directory name: " + self.asset_name)
            print(">>> Expected Manager file name: " + manager_file_name)
            print(">>> Expected Manager class name: " + manager_class_name)
            raise FrameworkException(str(e))

    def extract_stage(self, stage_name):
        """
        Gets the pipeline stage dictioanry containing the function and the manager class it resides in.
        :param stage_name: the name of the stage (e.g. "load_data")
        :return:  pipeline stage dictionary
        """
        asset_name = ''.join(x.capitalize() or '_' for x in self.asset_name.split('_'))  # CamelCase
        if asset_name not in AVAILABLE_STAGES:
            raise PipelineManagerException(
                "Missing decoration for your pipeline functions! Add '@pipeline' decorator above functions"
                " you want to use in your asset '{}'s Data Manager and Model Manager.".format(asset_name))

        if stage_name not in AVAILABLE_STAGES[asset_name]:
            # exists in one if the base classes
            if stage_name in AVAILABLE_STAGES[BASE_CLASS_NAME]:
                return AVAILABLE_STAGES[BASE_CLASS_NAME][stage_name]

            raise PipelineManagerException(
                "Function '{}' was not found in your asset! Add '@pipeline' decorator above your '{}' "
                "function if you want to use it in your pipeline.".format(stage_name, stage_name))

        return AVAILABLE_STAGES[asset_name][stage_name]

    def extract_manager_instance(self, manager_type):
        """
        Gets the instance of the manager - model_manager instance or data_manager instance
        :param manager_type: string "data_manager" or "model_manager"
        :return: model_manager instance or data_manager instance
        """
        if manager_type == 'data_manager':
            return self.data_manager_instance
        else:
            return self.model_manager_instance

    def run(self, *arguments):
        """
        Runs through the pipeline stages and passes relevant values between them
        :param arguments: input for the first stage in the pipeline, will be passed with *args
        :return: IOmanager of all the outputs to be stored
        """
        print(">>>>>> Running pipeline" + self.pipeline_name + "...")
        prev_stage_name = ''
        for stage_name in self.stages:
            start_time = time.strftime(TIME_FORMAT)
            print(">>>>>> Running stage: {}...".format(stage_name))

            stage = self.extract_stage(stage_name)
            if prev_stage_name:
                args = self.state[prev_stage_name]
            else:
                args = arguments

            # execute stage
            self.state[stage_name] = stage['function'](self.extract_manager_instance(stage['manager']), *args)

            if not isinstance(self.state[stage_name], tuple):
                self.state[stage_name] = (self.state[stage_name],)

            prev_stage_name = stage_name

            end_time = dt.datetime.strptime(
                time.strftime(TIME_FORMAT), TIME_FORMAT) - dt.datetime.strptime(start_time, TIME_FORMAT)
            print(">>>>>> It took me, {}.".format(end_time))

        print(">>>>>> Finished running pipeline.")
        return self.output_manager


class pipeline:
    def __init__(self, fn):
        self.fn = fn

    def __set_name__(self, owner, name):
        asset_name = owner.__name__

        manager_type = None
        for manager_type_key in MANAGER_TYPES:
            if MANAGER_TYPES[manager_type_key] in asset_name:
                manager_type = manager_type_key
        if manager_type is None:
            raise Exception("Wrong class name or placement of decorator! ('{}')".format(asset_name))

        asset_name = asset_name.replace('DataManager', '').replace('ModelManager', '')

        if asset_name not in AVAILABLE_STAGES:
            AVAILABLE_STAGES[asset_name] = {}

        if name in AVAILABLE_STAGES[asset_name]:
            raise Exception("Duplicate stage name '{}' for pipelines found in asset '{}'"
                            .format(asset_name, name))

        AVAILABLE_STAGES[asset_name][name] = {
            'function': self.fn,
            'manager': manager_type
        }

        return self.fn



from abc import ABCMeta, abstractmethod
from mlapp.config import settings
from mlapp.utils.exceptions.framework_exceptions import SkipServiceException


class WrapperInterface:
    __metaclass__ = ABCMeta

    def __init__(self):
        """
        Constructor for he Wrapper Interface.
        Will hold the handlers set up in the environment in the `self._handlers` in a dictionary by key name.
        Will hold a list containing the key names of handlers set up as the main ones.
        """
        self._handlers = {}
        self._main_handlers = []
        self.initialized = False

    @abstractmethod
    def init(self, handler_type):
        """
        Initialization, should be called once only
        Populates the `self._handlers` and `self_main_handlers` variables depending on the set environment
        :param handler_type: used for filtering services by the handler type
        """
        if not self.initialized:
            for service_name in settings.get('services', []):
                service_item = settings['services'][service_name]
                if 'type' not in service_item:
                    raise Exception("'{}' service is missing 'type' key, must be filled in config.py with"
                                    " the one of the following: database/file_storage/database/spark".format(service_name))
                if service_item['type'] == handler_type:
                    try:
                        self._handlers[service_name] = service_item['handler'](service_item.get('settings', {}))

                        # set it as main
                        if service_item.get('main', False):
                            self._main_handlers.append(service_name)

                    except SkipServiceException as e:
                        pass  # skipping this service
                    except Exception as e:
                        if service_item['handler'] is None:
                            raise Exception("'{}' service of type '{}' is missing a python library installation."
                                            .format(service_name, service_item.get('type')))
                        else:
                            raise e
            self.initialized = True

    def get(self, handler_name):
        """
        Get the handler instance by name
        :param handler_name: handler name string
        :return: Handler Instance
        """
        return self._handlers.get(handler_name)

    def empty(self):
        """
        Checks if there are configured handlers as "main"
        """
        return len(self._main_handlers) == 0



import numpy as np

# train configurations
advanced_regression_config = {
    "pipelines_configs": [
        {
            "data_settings": {
                "generate_forecast_data": "index_based",  # normal
                "generate_forecast_data_indices": [35, 42, 57]
            },
            "model_settings": {
                "variable_to_predict": "target"
            },
            "job_settings": {
                "asset_name": "advanced_regression",
                "pipeline": "forecast",
                "model_id": "latest"
            }
        }
    ]
}



import logging
import pandas as pd
import numpy as np
import datetime
import re
from sqlalchemy import create_engine, Column, Time, Integer, String, JSON
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from mlapp.handlers.databases.database_interface import DatabaseInterface


class Job(declarative_base()):
    __tablename__ = 'jobs'

    id = Column(String, primary_key=True)
    user = Column(String)
    data = Column(JSON)
    status_code = Column(Integer)
    status_msg = Column(String)
    created_at = Column(Time)
    updated_at = Column(Time)


class SQLAlchemyHandler(DatabaseInterface):
    def __init__(self, settings):
        """
        SQLAlchemyHandler constructor, inherits from DatabaseInterface, now also supports AWS IAM service authentication
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        self.conn = None
        self.engine = None
        self.token = None
        self.connections_parameters = settings
        self.query_placeholder = '%s'
        if settings.get('use_aws_IAM', False):
            try:
                import boto3
                rds = boto3.client(
                    'rds',
                    aws_access_key_id=self.connections_parameters["aws_access_key"],
                    aws_secret_access_key=self.connections_parameters["aws_secret_key"],
                )
                self.token = rds.generate_db_auth_token(
                    self.connections_parameters["hostname"],
                    self.connections_parameters["port"],
                    self.connections_parameters["user_id"],
                )
            except Exception as e:
                raise Exception('Missing AWS connection parameters, please provide proper aws_access_key and aws_secret and proper hostname, port and user_id')
        if self.connections_parameters == {}:
            logging.error('Missing database connection details. Please provide connection parameters to your database')



    def execute_query(self, query, params=None):
        """
        Executes a query in the database. If using string format `{i}` place holder, will replace to proper DB place holder (e.g. %s)
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used in the where clause of the query
        :return: result of query
        """
        self._connect_database()
        try:
            query, params = self._format_query(query, params)
            if params is None:
                params = []

            result = self.conn.execute(query, tuple(params))  # Syntax error in query
            self.conn.close()
            self.engine.dispose()
            if result.returns_rows:
                # we return a list of any row that was fetched
                return list(result)
            else:
                # number of rows matched by where criterion of an UPDATE or DELETE
                return result.rowcount
        except Exception as e:
            self._close_connection()
            raise e

    def insert_query(self, query, values):
        """
        Inserts values to your table in the database.
        :param query: please provide a query in the following form: "insert into table_name (column1, comlum2,,,) VALUES (%s,%s,,,)
        :param values: list of dictionaries, as the number of rows to insert i.e. [{value1,value2,,,}, {value11,value22,,,,},,,]
        """
        try:
            self._connect_database()
            self.conn.execute(query, values)
            self.conn.close()
            self.engine.dispose()
        except Exception as e:
            self._close_connection()
            raise e

    def insert_df(self, sql_table, df, batch_length=1000):
        """
        Inserts a DataFrame into a table in the database.
        :param sql_table: str - name of the table.
        :param df: DataFrame (Pandas, PySpark or other) - Matrix type DataFrame containing all values to insert.
        :param batch_length: int - length of the how many rows to insert from matrix at a time
        """
        try:
            num_batches = np.ceil(np.true_divide(len(df.index), batch_length))
            print('>>>>>>  Inserting {0} batches  >>>>  '.format(str(num_batches)))
            for i in range(int(num_batches)):
                self._connect_database()
                try:
                    cur_df = df.iloc[i * batch_length: min((i + 1) * batch_length, len(df.index))]
                    cur_df.to_sql(sql_table, con=self.conn, if_exists='append', index=False)
                except Exception as e:
                    print(e)
                    self.conn.close()
                    self.engine.dispose()
                    continue
                self.conn.close()
                self.engine.dispose()
        except Exception as e:
            self._close_connection()
            raise e

    def get_df(self, query, params=None):
        """
        Executes the query in the database, returns results as a DataFrame.
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used if necessary in query
        :return: result of query as a DataFrame
        """
        self._connect_database()
        try:
            query, params = self._format_query(query, params)
            if params is None:
                params = []
            result = self.conn.execute(query, tuple(params))
            df = pd.DataFrame(result.fetchall(), columns=result.keys())
            self.conn.close()
            self.engine.dispose()
            return df
        except Exception as e:
            self._close_connection()
            raise e

    def update_job_running(self, job_id):
        """
        Updates row in the table of jobs by the job_id to status `Running`
        Functionality of the Control Panel
        :param job_id: str - id of the job
        """
        self._connect_database()
        try:
            # create session
            session = sessionmaker(bind=self.engine)()

            # query by job id
            q = session.query(Job)
            q = q.filter(Job.id == job_id)

            # create or update
            records = q.all()
            if records:
                records[0].status_code = 1
                records[0].status_msg = 'running'
            else:
                session.add(Job(id=job_id, status_code=1, status_msg='running', user='DefaultUser',
                                created_at=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))

            # commit and close
            session.commit()
            session.close()
            self._close_connection()
        except Exception as e:
            self._close_connection()
            raise e

    def update_actuals(self, df, index_column='index',target_column='y_true'):
        # TODO: update to ORM (currently supports postgres only)
        query = 'update target set y_true = (case'
        columns = list(df)

        for index in columns:
            value = float(df[index])
            query += f" when index = '{str(index)}' then {str(value)}"
        query += ' end) where type=3 and index in (' + ','.join([f"'{x}'" for x in columns ]) + ')'
        self.execute_query(query)

    def get_model_predictions(self, model_id, prediction_type=3, from_date=None, to_date=None):
        # TODO: update to ORM (currently supports postgres only)
        # TODO: use from and to dates
        return self.get_df(
            query="select * from target where model_id = ? and type = ? ",
            params=[model_id, prediction_type]
        )

    ##################################################
    #                                                #
    #                Private methods                 #
    #                                                #
    #################################################


    def _connect_database(self):
        """
        Connect to the database.
        """
        try:
            self.engine = self._create_engine()
            self.conn = self.engine.connect()
        except Exception as e:
            raise e

    def _create_engine(self):
        """
        Create engine for the database.
        """
        if hasattr(self, 'connection_string'):
            return create_engine(self.connection_string, connect_args=self.connections_parameters['options'])
        else:
            raise NotImplementedError("Must initiate `self.connection_string` in your handler's `__init__` function!")

    def _close_connection(self):
        """
        Close connection to the database and dispose of engine.
        """
        try:
            if self.conn is not None and not self.conn.closed:
                self.conn.close()
                self.engine.dispose()
        except Exception as e:
            raise e

    def _format_query(self, query, params):
        """
        Formatting query and parameters. Replaces {number} with self.char (%s or other).
        This function handles tuples for IN clasue as well.
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used if necessary in query, if a tuple is passed for IN clause, need to wrap it in a list as well
        :return: result of query and new parameters
        """
        pattern = r'"?\{\d+\}"?'
        any_placeholders = re.findall(pattern, query)
        if len(any_placeholders) > 0 and len(any_placeholders) != len(params):
            raise Exception("Number of values and number of query placeholders mismatch")
        if len(any_placeholders) == 0:
            return query, params

        new_params = []
        for p in any_placeholders:
            param_index = int(re.findall(r'\d+', p)[0])
            param = params[param_index]
            new_placeholder = self.query_placeholder

            if isinstance(param, list) or isinstance(param, tuple):
                new_params.extend(param)
                new_placeholder = ', '.join([self.query_placeholder] * len(param))
                if not bool(re.findall(r'\(\s*' + new_placeholder, query)):
                    new_placeholder = '(' + new_placeholder + ')'
            else:
                new_params.extend([param])
            query = re.sub(r'"?\{' + str(param_index) + '\}"?', new_placeholder, query)
        return query, new_params

    def __del__(self):
        self._close_connection()





from azureml.core import Run
from mlapp.main import MLApp
from mlapp.handlers.wrappers.file_storage_wrapper import file_storage_instance
from mlapp.integrations.aml.utils.flow import parse_args, flow_setup, flow_postprocess
from mlapp.integrations.aml.utils.constants import PARSED_ARG_CONFIG, PARSED_ARG_INPUT_DIR, PARSED_ARG_OUTPUT_DIR
from mlapp.managers.flow_manager import FlowManager
from mlapp.integrations.aml.utils.run_class import load_config_from_string, tag_and_log_run, tag_and_log_outputs
from config import settings

# parsing arguments
parsed_args = parse_args()

# pre-processing
config = load_config_from_string(parsed_args[PARSED_ARG_CONFIG])
tag_and_log_run(config)

# current run identification
current_run = Run.get_context()

# init mlapp
MLApp(settings)

# flow setup
jobs_outputs = flow_setup(current_run.id, config, parsed_args[PARSED_ARG_INPUT_DIR])

# run config
_, output_ids, output_data = FlowManager(current_run.id, config, **jobs_outputs).run()

# post-processing
tag_and_log_outputs(output_ids)
file_storage_instance.postprocessing()
flow_postprocess(config, output_data, parsed_args[PARSED_ARG_OUTPUT_DIR])



spark_regression_config = {
    "pipelines_configs": [
        {
            # train config
            "data_settings": {
                "local_file_path": "data/diabetes.csv",
                "variable_to_predict": "target",  # column name of variable to predict
                # features handling configurations
                "data_handling": {
                    "features_for_train": [],  # leave empty for selecting all features
                    "set_features_index": [],  # leave empty for indexing by row index
                    # features to remove
                    "features_to_remove": [
                        "age", "sex", "tc", "ldl",
                        "age_sex", "sex_age",
                        "age_hdl", "hdl_age",
                        "sex_tc", "tc_sex",
                        "sex_ldl", "ldl_sex",
                        "bmi_hdl", "hdl_bmi",
                        "map_hdl", "hdl_map",
                        "tc_hdl", "hdl_tc",
                        "hdl_tch", "tch_hdl",
                        "hdl_ltg", "ltg_hdl",
                        "hdl_glu", "glu_hdl",
                    ],
                    "feature_remove_by_null_percentage": 0.3,
                    # features transformations and missing value strategies
                    "features_handling": {
                        "bmi": {"fillna": "mean", "transformation": ["log1p"]},
                        "hdl": {"fillna": "mean", "transformation": ["log1p"]},
                        "tch": {"fillna": "mean", "transformation": ["log1p"]},
                        "glu": {"fillna": "mean", "transformation": ["log1p"]}
                    },
                    "features_to_bin": [
                        {"name": "tch", "bins": [-0.05, 0.03]},
                        {"name": "hdl", "bins": [0.07]},
                    ],
                    "interactions": True,
                }
            },
            # model settings
            "model_settings": {
                "train_percent": 0.8,
                "variable_to_predict": "target"  # column name of variable to predict
            },
            # job settings
            "job_settings": {
                "asset_name": "spark_regression",
                "pipeline": "train"
            }
        }
    ]
}



import click, os
from mlapp.mlapp_cli.common.cli_utilities import set_env, create_file, create_directory
from mlapp.mlapp_cli.cli_help import cli_environment_help


@click.group("environment")
def commands():
    """
    ML App Environment Command

    """
    pass


@commands.command("init", help=cli_environment_help.get('init', 'init environment file'))
@click.argument("name", required=True, default='.env')
def init(name):
    try:
        if '.env' not in name:
            name += '.env'

        env_full_path = os.path.join(os.getcwd(), os.path.join('env', name))
        if not os.path.exists(env_full_path):

            # creates the env directory if not exists.
            create_directory(directory_name='env', include_init=False)

            # creates env file
            create_file(name, path='env')

            # set the new env file
            set_env(name)
        else:
            click.secho("ERROR: '" + name + "' file already exits.", fg='red')

    except Exception as e:
        click.secho("ERROR: Oops, something went wrong.", fg='red')


@commands.command("set", help=cli_environment_help.get('set', 'sets environment file'))
@click.argument("name", required=True)
def set(name):
    try:
        if '.env' not in name:
            name += '.env'

        # set the new env file
        set_env(name)

    except Exception as e:
        click.secho("ERROR: Oops, something went wrong.", fg='red')



from azureml.data.data_reference import DataReference
from azureml.pipeline.core import PipelineData

from mlapp.integrations.aml.utils.compute import get_or_create_compute_target
from mlapp.integrations.aml.utils.constants import OUTPUT_PATH_ON_COMPUTE, DATA_REFERENCE_NAME
from mlapp.integrations.aml.utils.pipeline import publish_pipeline_endpoint, create_mlapp_pipeline_step
from mlapp.integrations.aml.utils.runconfig import create_runconfig
import os


def run_script(ws, datastore, pipeline_name, instructions):
    pipeline_steps = []
    last_output = []

    for i in range(len(instructions)):
        compute_target = get_or_create_compute_target(
            ws,
            compute_name=instructions[i]['name'],
            vm_size=instructions[i].get('vm_size', 'STANDARD_D2_V2'),
            min_nodes=instructions[i].get('min_nodes', 0),
            max_nodes=instructions[i].get('max_nodes', 4),
            idle_sec=instructions[i].get('idle_seconds_before_scale_down', 120)
            )
        run_config = create_runconfig(compute_target)

        # input directory in datastore
        if len(last_output) == 0:
            input_dir = None
            # input_dir = DataReference(
            #     datastore=datastore,
            #     data_reference_name=DATA_REFERENCE_NAME + str(i),
            #     path_on_datastore="flows/",
            #     mode='download'
            # )
        else:
            input_dir = last_output

        # output directory in datastore
        output_dir = PipelineData(
            name=DATA_REFERENCE_NAME + str(i),
            datastore=datastore,
            output_path_on_compute=OUTPUT_PATH_ON_COMPUTE
        )

        # create pipeline step
        pipeline_step = create_mlapp_pipeline_step(
            compute_target, run_config,
            source_directory=os.getcwd(),
            entry_script=os.path.join("deployment", "aml_flow.py"),
            input_dir=input_dir,
            output_dir=output_dir,
            param_name='config' + str(i)
        )

        # add to pipeline
        pipeline_steps += pipeline_step

        # reference last output
        last_output.append(output_dir)

    publish_pipeline_endpoint(ws, pipeline_steps, pipeline_name)





import click
import os
from tabulate import tabulate
from mlapp.main import MLApp
from mlapp.mlapp_cli.common.services_options import add_services_options
from mlapp.mlapp_cli.common.cli_utilities import create_directory, validate_str, clean_spaces, \
    format_service_dictto_txt, write_to_file, check_for_service_uniqueness_name, get_env
from mlapp.mlapp_cli.cli_help import cli_services_help

# constants
EMPTY_INPUT = ""


@click.group("services")
def commands():
    """
    ML App Services Command

    """
    pass


@commands.command("add", help=cli_services_help.get('add', 'add service'))
@click.argument("service", required=True)
def add(service):
    # creates the env directory if not exists.
    create_directory(directory_name='env', include_init=False)

    # find env file
    try:
        env_filename = get_env()
    except Exception as e:
        click.secho("ERROR: Please run 'mlapp init' first or set environment file.", fg='red')
        return

    env_full_path = os.path.join(os.getcwd(), env_filename)
    credentials = {}
    service_key_cache = {}
    if os.path.exists(env_full_path):
        service_keys = add_services_options.get(service, False)
        if service_keys:

            # get service name from user
            while True:
                service_name = validate_str(input("Please name your service (to access the service in the code): "))
                is_unique = check_for_service_uniqueness_name(service_name, env_filename)
                if clean_spaces(service_name) != '' and is_unique:
                    break
                else:
                    click.secho("ERROR: Service name is required and must be unique, please try again.", fg='red')

            for key in service_keys:
                while True:
                    try:
                        body = service_keys[key]
                        new_key = service_name + '_' + key
                        if isinstance(body, str):
                            credentials[new_key] = body
                        else:
                            display_name = body.get('display_name', key)
                            short_description = body.get('short_description', '')

                            # check if this key depends on answer from previous key
                            prev_key = body.get("preceding_key", None)
                            if prev_key is not None and service_key_cache.get(prev_key) is not None:
                                if service_key_cache[prev_key] != body.get("preceding_key_response"):
                                    break

                            # create message to display on the terminal.
                            message = display_name + " (" + short_description + "): " if short_description != '' else key + ": "

                            # get user input
                            user_input = clean_spaces(input(message))

                            # check for default and required values
                            default_value = body.get('default', None)
                            is_required = body.get('required', False)

                            if default_value is not None and user_input == EMPTY_INPUT:
                                credentials[new_key] = default_value
                            elif user_input == EMPTY_INPUT and is_required:
                                raise Exception("'" + display_name + "' is required, please enter a valid value.")
                            else:

                                # validates user input
                                validations_methods = body.get('validations', [])
                                ans = True
                                for validate in validations_methods:
                                    ans &= validate(user_input)
                                    if not ans:
                                        raise Exception(body.get('error_msg', ''))

                                # transforms user input
                                transformations_methods = body.get('transformations', [])
                                for trans in transformations_methods:
                                    user_input = trans(user_input)

                                saved_value = user_input
                                body_values = body.get('values', False)
                                if body_values:
                                    body_values_keys = body['values'].keys()
                                    if user_input in body_values_keys:
                                        saved_value = body['values'][user_input]
                                        credentials[new_key] = saved_value
                                    else:
                                        raise Exception(body.get('error_msg', 'Oops something bad happened.'))
                                else:
                                    credentials[new_key] = user_input

                                if body.get('save_user_input', False):
                                    service_key_cache[key] = saved_value
                        break
                    except Exception as e:
                        if e == "":
                            click.secho("Invalid value please try again.", fg='red')
                        else:
                            click.secho(str(e), fg='red')

            comment = "\n# " + service_name.lower() + " " + service.lower() + " service\n"
            write_to_file(env_full_path, content=comment + format_service_dictto_txt(credentials))
            click.secho(
                "\nSuccess: " + service.capitalize() + " service was added to your project under the name " + service_name.upper() + ".\n",
                fg='green')
        else:
            click.secho("Service not found.", fg='red')
    else:
        # todo create a default env file?.
        click.secho('Env file not found, please create it before!', fg='red')


@commands.command("delete", help=cli_services_help.get('delete', 'delete service'))
@click.argument("service", required=True)
def delete(service):
    try:

        # find env file
        try:
            env_filename = get_env()
        except Exception as e:
            click.secho("ERROR: Please run 'mlapp init' first or set environment file.", fg='red')
            return

        env_full_path = os.path.join(os.getcwd(), env_filename)
        tmp_filename = "tmp_env_file_to_be_renamed.env"
        full_tmp_file_path = os.path.join(os.getcwd(), os.path.join('env', tmp_filename))
        is_service_name_to_del_found = False

        if os.path.exists(env_full_path):

            is_unique = check_for_service_uniqueness_name(service, env_filename)
            if is_unique:
                click.secho("ERROR: service not found.\nHint: type 'mlapp services' to see all available services.",
                            fg='red')
            else:
                with open(env_full_path) as f:
                    lines = f.readlines()

                    # creates new file with temporary name
                    with open(full_tmp_file_path, "w") as output:
                        for line in lines:
                            line_striped = line.strip()

                            if len(line_striped) == 0:
                                if is_service_name_to_del_found:
                                    continue
                                else:
                                    output.write(line)
                                    is_service_name_to_del_found = False
                                    continue

                            line_split = line_striped.split('_')

                            if len(line_split) > 0 and line_striped[0] != '#':
                                current_service_name = line_split[0]
                                # check if we should keep current service
                                if current_service_name.lower() != service.lower():
                                    output.write(line)
                                else:
                                    is_service_name_to_del_found = True
                            else:
                                if line_striped != '\n' and line_striped[0] == '#' and service.lower() in line_striped.lower():
                                    continue
                                else:
                                    output.write(line)

                # deletes original file
                os.remove(env_full_path)

                # renames tmp file to original file
                os.rename(full_tmp_file_path, env_full_path)

                click.secho(
                    "\nSuccess: " + service.upper() + " service was deleted from your project.\n", fg='green')

        else:
            click.secho('ERROR: environment file not found, please create it before!', fg='red')
    except Exception as e:
        raise e


@commands.command("show", help=cli_services_help.get('show', 'show services'))
def show():
    # find env file
    try:
        env_filename = get_env()
    except Exception as e:
        click.secho("ERROR: Please run 'mlapp init' first or set environment file.", fg='red')
        return

    env_full_path = os.path.join(os.getcwd(), env_filename)

    if os.path.exists(env_full_path):
        services = set()
        with open(env_full_path) as f:
            lines = f.readlines()
            for line in lines:
                line_striped = line.rstrip()
                line_split = line_striped.split('_')
                if len(line_split) > 1:
                    service_name = line_split[0].lstrip()
                    service_type = '_'.join(line_split[1:])
                    if service_name != '' and service_name[0] != '#' and MLApp.MLAPP_SERVICE_TYPE[1:] in service_type:
                        service_type = service_type.split('=')[1]
                        services.add(tuple([service_name, service_type]))

        if len(services):
            click.echo(tabulate(list(services), headers=['Service Name', 'Service Type']))
            click.echo()
        else:
            click.echo("No services registered")
    else:
        click.secho('Env file not found, please create it before!', fg='red')


@commands.command("show-types", help=cli_services_help.get('show_types', 'show services types'))
def show_types():
    services_keys = add_services_options.keys()
    click.echo('Available services types:')
    for key in services_keys:
        click.echo('[*] ' + key)

# if __name__ =='__main__':
#     add('postgres')



from mlapp.managers import ModelManager, pipeline
from sklearn.model_selection import train_test_split
from mlapp.utils.automl import AutoMLPandas


class CrashCourseModelManager(ModelManager):
    @pipeline
    def train_model(self, data):
        # prepare data for train
        X, y = data.drop(self.model_settings['target'], axis=1), data[self.model_settings['target']]
        X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=self.model_settings.get('train_percent'))

        # run auto ml
        result = AutoMLPandas('multi_class', **self.model_settings.get('auto_ml', {}))\
            .run(X_train, y_train, X_test, y_test)

        # print report
        result.print_report(full=False)

        # save results
        self.save_automl_result(result)

    @pipeline
    def forecast(self, data):
        # load the model
        model = self.get_automl_result().get_best_model()

        # predict
        predictions = model.predict(data)

        # store the predictions
        self.add_predictions(data.index,  predictions, [], prediction_type='forecast')

    @pipeline
    def refit(self, data):
        raise NotImplementedError()



import os, click
from mlapp.mlapp_cli.common.cli_utilities import create_file, create_directory, set_env, get_available_port
from mlapp.mlapp_cli.common.files import docker_compose_file, env_file, default_config_file, vue_env_config_file, init_sql_file
from mlapp.mlapp_cli.cli_help import cli_cp_help


@click.group("cp")
def commands():
    """
    MLApp Control Panel Command

    Use it to setup and run ML App Control Panel locally on your machine.

    """
    pass


@commands.command("setup", help=cli_cp_help.get('setup', 'setup cp on your machine'))
def setup():
    # creates the env directory if not exists.
    create_directory(directory_name='env', include_init=False)

    # creates the deployment directory if not exists.
    create_directory(directory_name='deployment', include_init=False)

    docker_compose_file_content = docker_compose_file
    init_sql_file_content = init_sql_file
    env_file_content = env_file

    # creates files
    create_file(file_name='docker-compose.yaml', path='deployment', content=docker_compose_file_content)
    # create_file(file_name='init.sql', path='deployment', content=init_sql_file_content)
    create_file(file_name='env-config.js', path='deployment', content=vue_env_config_file)

    create_file(file_name='.env', path='env', content=env_file_content)

    # edit content of config.py file, set env new file
    default_env_filename = ''
    config_file_content = default_config_file.replace("<FILENAME>", default_env_filename)
    create_file(file_name='config.py', content=config_file_content)


@commands.command("start", help=cli_cp_help.get('start', 'start cp'))
def start():
    is_config = os.path.exists(os.path.join(os.getcwd(), 'config.py'))
    is_yaml = os.path.exists(os.path.join(os.getcwd(), 'deployment/docker-compose.yaml'))
    is_env = os.path.exists(os.path.join(os.getcwd(), 'env/.env'))

    if is_config and is_env and is_yaml:

        # find available port
        # init_ui_port = 8081
        # ui_port = get_available_port(init_ui_port)
        #
        # if ui_port != init_ui_port:
        #     click.echo('Warning: port ' + str(init_ui_port) +' already in use, using ' + str(ui_port) + ' instead.')

        # set env file
        set_env(env_filename='.env')

        # start command
        cmd = "docker-compose up"

        # set cwd to deployment folder
        os.chdir(os.path.join(os.getcwd(), "deployment"))

        os.system(cmd)

        # set cwd back to root
        os.chdir("../")
    else:
        click.secho("ERROR: Please run 'mlapp cp setup' command before starting.", fg='red')


@commands.command("stop", help=cli_cp_help.get('stop', 'stop cp'))
def stop():
    # set env file
    set_env(env_filename='')

    # set cwd to deployment folder
    os.chdir(os.path.join(os.getcwd(), "deployment"))

    cmd = "docker-compose down"
    os.system(cmd)

    # set cwd back to root
    os.chdir("../")



from azureml.core import ComputeTarget
from azureml.core.compute import AmlCompute


def get_or_create_compute_target(workspace, compute_name, vm_size='STANDARD_D2_V2', min_nodes=0, max_nodes=4,
                                 idle_sec=120):
    if compute_name in workspace.compute_targets:
        compute_target = workspace.compute_targets[compute_name]
        if compute_target and type(compute_target) is AmlCompute:
            print('Found compute target: ' + compute_name)
    else:
        print('Creating a new compute target...')
        provisioning_config = AmlCompute.provisioning_configuration(vm_size=vm_size,  # STANDARD_NC6 is GPU-enabled
                                                                    min_nodes=min_nodes,
                                                                    max_nodes=max_nodes,
                                                                    idle_seconds_before_scaledown=idle_sec)
        # create the compute target
        compute_target = ComputeTarget.create(
            workspace, compute_name, provisioning_config)

        # Can poll for a minimum number of nodes and for a specific timeout.
        # If no min node count is provided it will use the scale settings for the cluster
        compute_target.wait_for_completion(show_output=True)

        # For a more detailed view of current cluster status, use the 'status' property
        print(compute_target.status.serialize())

    return compute_target



from mlapp.handlers.databases.sql_alchemy_handler import SQLAlchemyHandler
import ssl


class PostgresHandler(SQLAlchemyHandler):
    def __init__(self, settings):
        """
        PostgresHandler init with it's special connection string. Also supports for AWS IAM connection token
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        # add ssl
        if settings.get('ssl'):
            ssl_context = ssl.SSLContext()
            # ssl_context.verify_mode = ssl.CERT_REQUIRED
            # ssl_context.load_verify_locations('./certs/postgres.pem')
            settings['options'] = {
                'ssl_context': ssl_context
            }

        super(PostgresHandler, self).__init__(settings)

        password = self.connections_parameters['password'] if self.token is None else self.token

        self.connection_string = 'postgresql+pg8000://{0}:{1}@{2}:{3}/{4}'.format(
            self.connections_parameters['user_id'],
            password,
            self.connections_parameters['hostname'],
            str(self.connections_parameters['port']),
            self.connections_parameters['database_name'])




from mlapp.handlers.file_storages.file_storage_interface import FileStorageInterface
from azureml.core.dataset import Dataset
from azureml.core import Run
import pandas as pd
from azureml.exceptions import RunEnvironmentException
from mlapp.utils.exceptions.framework_exceptions import SkipServiceException


class AmlDatasetHandler(FileStorageInterface):
    def __init__(self, settings):
        super(AmlDatasetHandler, self).__init__()

    def download_file(self, dataset_name, dataset_type='pandas', *args, **kwargs):
        """
        Downloads file from file storage
        :param dataset_name: name of the dataset
        :param dataset_type: name of the dataset type
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        try:
            run = Run.get_context(allow_offline=False)
        except RunEnvironmentException as e:
            raise SkipServiceException('Skip AmlModelStorageHandler handler')
        ws = run.experiment.workspace
        ds = Dataset.get_by_name(workspace=ws, name=dataset_name)  # Get a Dataset by name
        if dataset_type == 'spark':
            df = ds.to_spark_dataframe()  # Load a Tabular Dataset into pandas DataFrame
        else:
            df = ds.to_pandas_dataframe()  # Load a Tabular Dataset into pandas DataFrame
        return df

    def upload_file(self, dataset: pd.DataFrame, dataset_name: str, dataset_type: str='pandas', *args, **kwargs):
        """
        Uploads file to file storage
        :param dataset: dataset object
        :param dataset_name: name of the dataset
        :param dataset_type: name of the dataset type
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        run = Run.get_context()
        ws = run.experiment.workspace
        if dataset_type == 'pandas':
            ds = Dataset.from_pandas_dataframe(dataframe=dataset)  # Load a Tabular Dataset into pandas DataFrame
        else:
            raise Exception("ERROR: only pandas DataFrame is supported ath the moment.")

        ds.register(ws, name=dataset_name)

    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        raise Exception("ERROR: Azure Machine Lreaning Dataset class not supported this method.")



from mlapp.handlers.databases.database_interface import DatabaseInterface
from mlapp.handlers.file_storages.file_storage_interface import FileStorageInterface
from mlapp.handlers.message_queues.message_queue_interface import MessageQueueInterface
from mlapp.handlers.spark.spark_interface import SparkInterface
from mlapp.handlers.wrappers.database_wrapper import database_instance
from mlapp.handlers.wrappers.file_storage_wrapper import file_storage_instance
from mlapp.handlers.wrappers.message_queue_wrapper import message_queue_instance
from mlapp.handlers.wrappers.spark_wrapper import spark_instance


def _get_handler(instance_type, handler_name):
    handler = instance_type.get(handler_name)
    if handler is not None:
        return handler
    raise Exception("Handler %s not found" % str(handler_name))


def database_handler(handler_name) -> DatabaseInterface:
    return _get_handler(database_instance, handler_name)


def file_storage_handler(handler_name) -> FileStorageInterface:
    return _get_handler(file_storage_instance, handler_name)


def message_queue_handler(handler_name) -> MessageQueueInterface:
    return _get_handler(message_queue_instance, handler_name)


def spark_handler(handler_name) -> SparkInterface:
    return _get_handler(spark_instance, handler_name)




import click
import os

# mlapp CLI
import mlapp.mlapp_cli.assets
import mlapp.mlapp_cli.boilerplates
import mlapp.mlapp_cli.services
import mlapp.mlapp_cli.environment
import mlapp.mlapp_cli.cp
from mlapp.mlapp_cli.init import init_command
from mlapp.mlapp_cli.common.files import gitignore_file, dockerignore_file
from mlapp.mlapp_cli.common.cli_utilities import create_file
from mlapp.version import VERSION
try:
    from mlapp.mlapp_cli.cli_test_env import check_env_test
except ModuleNotFoundError:
    check_env_test = None

try:
    from mlapp.integrations.aml.cli import commands as aml_commands
except:
    aml_commands = None

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

if check_env_test and callable(check_env_test):
    check_env_test()

# ASCII - standard font
LOGO = r""" 
IBM Services Framework for
███    ███ ██       █████  ██████  ██████  
████  ████ ██      ██   ██ ██   ██ ██   ██ 
██ ████ ██ ██      ███████ ██████  ██████  
██  ██  ██ ██      ██   ██ ██      ██      
██      ██ ███████ ██   ██ ██      ██       {}
""".format(VERSION)


@click.group(name="mlapp", context_settings=CONTEXT_SETTINGS)
@click.version_option("", "--version", "-V", "-v", help="Show version and exit", prog_name="", message=LOGO)
def cli():
    pass


@cli.command(help="Creates an initial project structure")
@click.option("-cp", "--control-panel", is_flag=True, default=False,
              help="Flag that includes the ML control panel in your project.")
@click.option("-aml", "--azure-machine-learning", is_flag=True, default=False, hidden=aml_commands is None,
              help="Flag that includes the AML setup files in your project.")
@click.option("-g", "--gitignore", is_flag=True, default=False,
              help="Flag that disables addition of a .gitignore file into your project.")
@click.option("-d", "--dockerignore", is_flag=True, default=False,
              help="Flag that disables addition of a .dockerignore file into your project.")
@click.option("-f", "--force", is_flag=True, default=False,
              help="Flag force init if project folder is not empty.")
def init(control_panel, azure_machine_learning, gitignore, dockerignore, force):
    init_command(control_panel, azure_machine_learning, gitignore, dockerignore, force)


@cli.command(help="Use to create ML App recommended '.gitigonre' file.")
@click.option("-f", "--force", is_flag=True, default=False,
              help="Flag force will override existing '.gitignore' file.")
def create_gitignore(force):
    if not os.path.exists(os.path.join(os.getcwd(), '.gitignore')) or force:
        create_file('.gitignore', content=gitignore_file)
    else:
        click.echo("Error: '.gitignore' already exists.")


@cli.command(help="Use to create ML App recommended '.dockerignore' file.")
@click.option("-f", "--force", is_flag=True, default=False,
              help="Flag force will override existing '.dockerignore' file.")
def create_dockerignore(force):
    if not os.path.exists(os.path.join(os.getcwd(), '.dockerignore')) or force:
        create_file('.dockerignore', content=dockerignore_file)
    else:
        click.echo("Error: '.dockerignore' already exists.")


cli.add_command(mlapp.mlapp_cli.assets.commands)
cli.add_command(mlapp.mlapp_cli.boilerplates.commands)
cli.add_command(mlapp.mlapp_cli.services.commands)
cli.add_command(mlapp.mlapp_cli.environment.commands)
cli.add_command(mlapp.mlapp_cli.cp.commands)

# adds integrations CLI's
if aml_commands is not None:
    cli.add_command(aml_commands)

if __name__ == "__main__":
    pass



from sklearn.metrics import confusion_matrix
from sklearn.utils.multiclass import unique_labels
import numpy as np
import matplotlib.pyplot as plt


def _generate_confusion_matrix(y_true, y_pred, normalize):
    # Utility to generate the data used for the confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
    return cm


def _plot_confusion_matrix(cm, ax, classes, title, normalize, cmap):
    # Utility to generate one confusion matrix image
    im = ax.matshow(cm, interpolation='nearest', cmap=cmap)
    ax.xaxis.set_ticks_position('bottom')
    ax.figure.colorbar(im, ax=ax)

    # We want to show all ticks...
    ax.set(xticklabels=['']+classes.tolist(), yticklabels=['']+classes.tolist())
    ax.set_aspect('auto')
    ax.set_title(title, fontdict={'fontsize': 8, 'fontweight': 'medium'})
    ax.set_xlabel('Predicted label', fontsize=6)
    ax.set_ylabel('True label', fontsize=6)  # relative to plt.rcParams['font.size']

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
    ax.margins(0.2)  # added to fix graph formatting
    # Loop over data dimensions and create text annotations.
    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(
                j, i, format(cm[i, j], fmt), ha="center", va="center", color="white" if cm[i, j] > thresh else "black")
    return ax


def plot_multiple_confusion_matrix(y_hat_train, y_train, y_hat_test, y_test):
    """
    This function will plot the confusion matrices for the train set and the test set, both as number and as proportion.
    :param model: trained instances of the model
    :param y_hat_train: pandas dataframe of the train set
    :param y_train: pandas series of the target related to the train set
    :param y_hat_test: pandas dataframe of the test set
    :param y_test: pandas series of the target related to the train set
    :return: One figure containing 4 images: confusion matrix for train set as number, confusion matrix for train set as
    frequencies, confusion matrix for test set as number, confusion matrix for test set as frequencies
    """
    fig, axes = plt.subplots(2, 2)

    titles = ['Train Confusion Matrix, without Normalization', 'Normalized Train Confusion Matrix',
              'Test Confusion Matrix, without Normalization', 'Normalized Test Confusion Matrix']

    norm_cm_train = _generate_confusion_matrix(y_train, y_hat_train, normalize=True)
    no_norm_cm_train = _generate_confusion_matrix(y_train, y_hat_train, normalize=False)
    norm_cm_test = _generate_confusion_matrix(y_test, y_hat_test, normalize=True)
    no_norm_cm_test = _generate_confusion_matrix(y_test, y_hat_test, normalize=False)

    train_classes = unique_labels(y_train, y_hat_train)
    test_classes = unique_labels(y_test, y_hat_test)

    axes[0, 0] = _plot_confusion_matrix(no_norm_cm_train, axes[0, 0], train_classes, titles[0], False, plt.cm.Blues)
    axes[0, 1] = _plot_confusion_matrix(norm_cm_train, axes[0, 1], train_classes, titles[1], True, plt.cm.Blues)
    axes[1, 0] = _plot_confusion_matrix(no_norm_cm_test, axes[1, 0], test_classes, titles[2], False, plt.cm.Blues)
    axes[1, 1] = _plot_confusion_matrix(norm_cm_test, axes[1, 1], test_classes, titles[3], True, plt.cm.Blues)

    fig.tight_layout()
    return fig



cli_aml_help = {
    "setup": \
'''Usage:

`mlapp aml setup` - setup Azure Machine Learning deployment files in your project.
''',
    "publish_pipeline": \
'''Usage:

`mlapp aml publish-pipeline [PIPELINE_NAME] [COMPUTE_TRAGET]` - set the pipeline name and the correct compute target.

for further options look on the options section below.
''',
    "deploy_model": \
'''Usage:

`mlapp aml deploy-model [EXPERIMENT_NAME] [ASSET_NAME]` - set your asset name.

for further options look on the options section below.
''',
    "add-mlapp-env": \
'''Usage:

`mlapp aml create-mlapp-env` - create a new python environment containing the ML App library in AzureML. 

for further options look on the options section below.
'''
}


from mlapp.managers import ModelManager, pipeline
import numpy as np
from pyspark.ml.feature import VectorAssembler
from mlapp.utils.automl import AutoMLSpark


class SparkClassificationModelManager(ModelManager):
    @pipeline
    def train_model(self, data):
        # preparation for train
        variable_to_predict = self.model_settings['variable_to_predict']

        # split the data to test and train
        train, test = data.randomSplit([0.9, 0.1], seed=12345)

        # TODO: scale and save scaler
        # scaler = MinMaxScaler(inputCol="features", outputCol='scaledFeatures')
        # scaler_model = scaler.fit(train_vec)
        #
        # train_scaled = scaler_model.transform(train_vec)
        # test_scaled = scaler_model.transform(test_vec)

        # run auto ml
        fs = self.model_settings.get('auto_ml', {}).get('feature_selection')
        result = AutoMLSpark('binary', **self.model_settings.get('auto_ml', {}).get('binary', {}),
                             feature_selection=fs).run(train, test, variable_to_predict, cv=3)

        # print report
        result.print_report()

        # adding objects
        self.save_automl_result(result, obj_type='pyspark')

        # adding predictions
        train_predictions = result.get_train_predictions().select('prediction', variable_to_predict).toPandas()
        test_predictions = result.get_test_predictions().select('prediction', variable_to_predict).toPandas()

        self.add_predictions(train_predictions.index, train_predictions['prediction'],
                             train_predictions[variable_to_predict], 'TRAIN')
        self.add_predictions(test_predictions.index, test_predictions['prediction'],
                             test_predictions[variable_to_predict], 'TEST')

    @pipeline
    def forecast(self, data_df):
        result = self.get_automl_result()

        # get best model
        model = result.get_best_model()

        # get selected features from feature selection algorithm
        selected_features_names = result.get_selected_features()

        # filter data according to selected features
        filtered_data = data_df.select(selected_features_names)

        # assemble data for train
        assembler = VectorAssembler(inputCols=selected_features_names, outputCol='features')
        data_df = assembler.transform(filtered_data).select('features')

        # prediction
        predictions = model.transform(data_df)
        num_of_predictions = predictions.count()

        # adding prediction to results
        self.add_predictions(
            np.arange(num_of_predictions),                                  # index array
            predictions.select('prediction').toPandas()['prediction'],      # y_hat
            np.array([np.nan] * num_of_predictions),                        # y_true (empty array)
            'forecast')                                                     # prediction type

    @pipeline
    def refit(self, data):
        raise NotImplementedError()



import os
import traceback
import pandas as pd
from config import settings
from mlapp import MLApp
from mlapp.integrations.aml.utils.deploy import preprocess_deployment, get_predictions_path, insert_model_id
from mlapp.integrations.aml.utils.run_class import load_config_from_string


def init():
    global run_id
    model_path = os.path.join(os.getcwd(), os.getenv('AZUREML_MODEL_DIR'))
    os.chdir('<LOCAL_ROOT_DIRECTORY_NAME>')
    output_dir = os.path.join(os.getcwd(), 'output')
    os.makedirs(output_dir, exist_ok=True)
    run_id = preprocess_deployment(model_path)


def run(raw_config):
    try:
        # loading config from string into dict object
        config = load_config_from_string(raw_config)

        # inserting deployed model id to config
        insert_model_id(config, run_id)

        # running config
        mlapp = MLApp(settings)
        _, ids, outputs = mlapp.run_flow_from_config(config)

        # loading predictions and returning it
        file_path = get_predictions_path(ids)
        predictions_df = pd.read_csv(file_path)
        return predictions_df.to_json(orient='records')

    except Exception as e:
        # printing error
        traceback.print_exc()
        return str(e)



from mlapp.managers import ModelManager, pipeline
import pandas as pd
from numpy import random
from sklearn.model_selection import KFold, train_test_split
from mlapp.utils.automl import AutoMLPandas


class AdvancedRegressionModelManager(ModelManager):
    @pipeline
    def train_model(self, data):
        """
        This function trains the model with the various algorithms and selects the best one.
        :param data: pandas DataFrame that contains X, y for training the model
        :return None
        """
        # preparation for train
        variable_to_predict = self.model_settings['variable_to_predict']
        X_train, X_test, y_train, y_test = train_test_split(
            data.drop(variable_to_predict, axis=1), data[variable_to_predict],
            train_size=self.model_settings.get('train_percent', 0.8), random_state=random.seed(0))

        # auto ml
        cv_splitter = KFold(n_splits=5, shuffle=True, random_state=random.seed(0))
        fs = self.model_settings.get('auto_ml', {}).get('feature_selection')

        # linear models
        linear_result = AutoMLPandas(
            'linear', feature_selection=fs, **self.model_settings.get('auto_ml', {}).get('estimators', {}).get(
                'linear', {})).run(X_train, y_train, X_test, y_test, cv=cv_splitter)

        # non linear models
        non_linear_result = AutoMLPandas(
            'non_linear', feature_selection=fs, **self.model_settings.get('auto_ml', {}).get('estimators', {}).get(
                'non_linear', {})).run(X_train, y_train, X_test, y_test, cv=cv_splitter)

        # merge results
        final_result = linear_result
        final_result.merge(non_linear_result)
        
        # adding best model
        self.save_automl_result(final_result)

        # adding predictions
        self.add_predictions(X_train.index, final_result.get_train_predictions(), y_train, 'TRAIN')
        self.add_predictions(X_test.index, final_result.get_test_predictions(), y_test, 'TEST')

    @pipeline
    def forecast(self, data_df):
        result = self.get_automl_result()

        # get best model
        model = result.get_best_model()

        # get selected features from feature selection algorithm
        selected_features_names = result.get_selected_features()

        # filter data according to selected features
        filtered_data = data_df[selected_features_names]

        # predicting
        predictions = model.predict(filtered_data)
        predictions_df = pd.DataFrame(data=predictions, columns=['y_hat'], index=filtered_data.index)
        predictions_df['y_true'] = None      # consistency with estimations (possibly add in the actual in the future)
        predictions_df['type'] = 3    # type 3 = forecast
        self.add_predictions(filtered_data.index, predictions, [], 'forecast')

    @pipeline
    def refit(self, data):
        raise NotImplementedError()





import numpy as np

# train configurations
classification_config = {
    "pipelines_configs": [
        {
            "model_settings": {
                "variable_to_predict": "answer",
                "down_sample_method": {
                    "flag": False,
                    "n_samples": 100,
                    "seed": 1500
                },
                "train_percent": 0.8,
                "predict_proba_threshold": [0.05, 0.5, 0.95],
                "auto_ml": {
                    "feature_selection": [
                        {
                            "method": "SelectKBest",
                            "params": {
                                'k': 7,
                                'score_func': 'chi2'
                            }
                        }
                    ],
                    "binary": {
                        "models": ["XGBoostClassifier", "Logistic", "ExtraTreeClassifier"],
                        "fixed_params": {
                            'Logistic': {
                                'solver': 'liblinear',
                                'max_iter': 500
                            },
                            'ExtraTreeClassifier': {
                                'min_samples_leaf': 4,
                                'max_depth': 10,
                                'class_weight': 'balanced'
                            }
                        },
                        "hyper_params": {
                            'Logistic': {
                                'C': 'list(np.linspace(0.01, 4, 15))',
                                'penalty': ['l1', 'l2'],
                                'class_weight': ['balanced', None],
                                'fit_intercept': [True],
                            },
                            'XGBoostClassifier': {
                                'max_depth': [3, 7, 10],
                                'learning_rate': 'list(np.linspace(0.001, 0.02, 3))',
                                'n_estimators': [300, 500, 700],
                                'min_child_weight': [3, 10],
                            },
                            'ExtraTreeClassifier': {
                                'n_estimators': [10, 50, 200],
                                'min_samples_split': [2, 10],
                            },
                        }
                    }
                }
            },
            "job_settings": {
                "asset_name": "classification",
                "pipeline": "reuse_features_and_train",
                "reuse_features_id": "latest",
                "data_id": "latest"
            }
        }
    ]
}



import logging
import boto3
from botocore.exceptions import ClientError
from mlapp.handlers.file_storages.file_storage_interface import FileStorageInterface


class BotoStorageHandler(FileStorageInterface):
    def __init__(self, settings):
        """
        Initializes the BotoStorageHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(BotoStorageHandler, self).__init__()
        configuration = settings

        if not configuration:
            logging.error('Configuration should be added to the file. Key should be "boto"')

        try:
            endpoint_url = configuration['endpoint'] if configuration['endpoint'] != "" else None
            self.botoClient = boto3.client(
                service_name='s3',
                region_name=None,
                api_version=None,
                use_ssl=True,
                verify=None,
                endpoint_url=endpoint_url,
                aws_access_key_id=configuration['accessKeyId'],
                aws_secret_access_key=configuration['secretKey'],
                aws_session_token=None,
                config=None)

        except KeyError as e:
            logging.error("Missing parameter in file storage config %s" % str(e))

    def download_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Downloads file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        self.botoClient.download_file(bucket_name, object_name, file_path)

    def upload_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Uploads file to file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        try:
            self.botoClient.create_bucket(Bucket=bucket_name)
        except Exception as e:
            pass
        try:
            self.botoClient.upload_file(file_path, bucket_name, object_name)
            logging.info('File storage: file %s successfully uploaded', object_name)
        except ClientError as e:
            logging.error(e)

    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        try:
            files_names = []
            response = self.botoClient.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
            for file in response['Contents']:
                files_names.append(file['Key'])
            return files_names
        except Exception as e:
            logging.error(e)



from azureml.core import Experiment, Workspace
from azureml.pipeline.core import PipelineParameter, Pipeline, PublishedPipeline
from azureml.pipeline.steps import PythonScriptStep, DatabricksStep

from mlapp.integrations.aml.utils.constants import INPUT_DATA_ARGUMENT, OUTPUT_DATA_ARGUMENT


def create_mlapp_pipeline_step(compute_target, run_config, source_directory, entry_script,
                                   input_dir=None, output_dir=None, param_name='config'):
    # adding pipeline parameter
    pipeline_param = PipelineParameter(
        name=param_name,
        default_value='{"pipelines_configs": [{"data_settings": {}, "model_settings": {}, '
                      '"job_settings":{"pipeline": "", "asset_name": ""}}]}')

    arguments = [("--" + param_name), pipeline_param]
    if input_dir is not None or output_dir is not None:
        if input_dir is not None and not isinstance(input_dir, list):
            input_dir = [input_dir]

        if output_dir is not None and not isinstance(output_dir, list):
            output_dir = [output_dir]

        for dir_type in [{'value': input_dir, 'str': INPUT_DATA_ARGUMENT},
                         {'value': output_dir, 'str': OUTPUT_DATA_ARGUMENT}]:
            if dir_type['value'] is not None:
                for i in range(len(dir_type['value'])):
                    arguments.append('--' + dir_type['str'] + str(i))
                    arguments.append(dir_type['value'][i])

    # pipeline step
    return [PythonScriptStep(
        script_name=entry_script,
        arguments=arguments,
        compute_target=compute_target,
        runconfig=run_config,
        allow_reuse=False,
        inputs=input_dir, outputs=output_dir,
        source_directory=source_directory
    )]


def create_databricks_pipeline_step(compute_target, source_directory, entry_script, num_workers=1):
    pipeline_param = PipelineParameter(
        name="config",
        default_value='{"pipelines_configs": [{"data_settings": {}, "model_settings": {}, '
                      '"job_settings":{"pipeline": "", "asset_name": ""}}]}')

    # input directory in datastore
    # input_dir = DataReference(datastore=datastore, data_reference_name="input_dir", path_on_datastore="configs",
    #     mode='download')

    # output directory in datastore
    # output_dir = PipelineData(name="output_dir", datastore=datastore, output_path_on_compute='results')

    dbStep = DatabricksStep(
        'databricksstep',
        inputs=[pipeline_param],    # ,input_dir],
        # outputs=[output_dir],
        num_workers=num_workers,
        source_directory=source_directory,
        python_script_path=entry_script,
        compute_target=compute_target,
        allow_reuse=False
    )
    return [dbStep]


def publish_pipeline_endpoint(workspace, steps, name='mlapp', description='My Published ML App', version='1.0'):
    # publish pipeline
    mlapp_pipeline = Pipeline(workspace=workspace, steps=steps)
    mlapp_pipeline.publish(name=name, description=description, version=version)


def run_pipeline_endpoint(workspace: Workspace, pipeline_endpoint_id, experiment_name, config_str,
                          pipeline_version=None):
    pipeline_endpoint_by_name = PublishedPipeline.get(workspace=workspace, id=pipeline_endpoint_id)

    run_id = pipeline_endpoint_by_name.submit(
        workspace=workspace,
        experiment_name=experiment_name,
        pipeline_parameters={
            'config': config_str
        })
    print(run_id)

    # option 2 with active directory
    # aad_token = InteractiveLoginAuthentication(tenant_id="my-tenant-id")
    # rest_endpoint = ''
    # response = requests.post(rest_endpoint, headers=aad_token, json={"ExperimentName": "PipelineEndpointExperiment",
    #                          "RunSource": "API", "ParameterAssignments": config_str})


def run_pipeline_steps(workspace, steps, experiment_name):
    mlapp_pipeline = Pipeline(workspace=workspace, steps=steps)
    pipeline_run = Experiment(workspace, experiment_name).submit(mlapp_pipeline)
    pipeline_run.wait_for_completion()



from mlapp.managers import DataManager, pipeline
import pandas as pd
import os
import numpy as np


class BasicRegressionDataManager(DataManager):
    # -------------------------------------- train methods -------------------------------------------
    @pipeline
    def load_train_data(self, *args):
        local_path = os.path.join(os.getcwd(), self.data_settings.get('local_file_path'))
        data = self._load_data_from_file(local_path)
        return data

    @pipeline
    def clean_train_data(self, data):
        data_handling = self.data_settings.get('data_handling', {})
        null_percentage = data_handling.get("feature_remove_by_null_percentage", 0.5)
        data = data.loc[:, data.isnull().mean() < null_percentage]
        missing_values = data.mean(axis=0).to_dict()
        data = data.fillna(missing_values)
        self.save_metadata('missing_values', missing_values)
        return data

    @pipeline
    def transform_train_data(self, data):
        return self._transform_data(data)

    # ------------------------------------- forecast methods -----------------------------------------
    @pipeline
    def load_forecast_data(self, *args):
        local_path = os.path.join(os.getcwd(), self.data_settings.get('local_file_path'))
        data = self._load_data_from_file(local_path)
        
        # creating predict data for example
        data.drop([self.model_settings['variable_to_predict']], axis=1, inplace=True)
        data = pd.DataFrame(data=np.random.normal(data.mean().mean(), data.std().mean(), data.shape)[0:10],
                            columns=data.columns)
        return data

    @pipeline
    def clean_forecast_data(self, data):
        missing_values = self.get_metadata('missing_values', {})
        data = data.fillna(missing_values)
        return data

    @pipeline
    def transform_forecast_data(self, data):
        return self._transform_data(data)

    # ------------------------- load helper function for train/forecast ------------------------------
    @staticmethod
    def _load_data_from_file(path):
        return pd.read_csv(path, encoding='ISO-8859-1')

    @pipeline
    def load_target_data(self, *args):
        raise NotImplementedError()

    def _transform_data(self, data):
        data_handling = self.data_settings.get('data_handling', {})

        # dropping features
        features_to_remove = data_handling.get('features_to_remove', [])
        if len(features_to_remove) > 0:
            data = data.drop([feature for feature in features_to_remove if feature in data.columns], axis=1)
        return data




